#!/usr/bin/env python2.3

import os, sys, os.path as osp, popen2, signal
from os.path import join as opj, expanduser as xu
from commands import getoutput as cmd
from twisted.python import usage
from glob import glob


TRIAL_ARGS = '-v quotient.test atop.test nevow.test'

def sh(cmdstr):
    status = os.system(cmdstr)
    if status != 0: 
        raise RuntimeError, "%s failed with exit code %s" % (cmdstr, status)

def cdAndRun(dir, f, *a, **kw):
    """guarantees as much as possible that after we cd and run f, that we will return to the
    original working directory """
    wd = osp.abspath(os.getcwd())
    try:
        os.chdir(dir)
        f(*a, **kw)
    finally:
        os.chdir(wd)


class Package(object):
    pkgdir = None
    def __init__(self, name, url):
        """name should be the name of the package that will appear before the version number of the tarball
        i.e. 'Lupy' or 'spambayes'
        """
        self.name, self.url = name, url

    def _globPkgDir(self):
        return glob("%s/%s*t*gz" % (self.pkgdir, self.name)) 
    pkgDirGlob = property(_globPkgDir)
    
    def _maybeGet(self):
        print "self.pkgDirGlob: %s, self.url: %s" % (self.pkgDirGlob, self.url)
        if self.pkgDirGlob or not self.url:
            return
        cdAndRun(self.pkgdir, sh, 'wget %s' % self.url)
            
    def _untarAndInstall(self):
        tarball = self.pkgDirGlob[0]
        untarDirName = cmd("tar ztf %s" % tarball).split()[0].replace('/', '')  # UGRH! this would make the baby jesus cry

        sh('tar zxvf %s' % tarball)
        sh('chmod -R ugo+rw %s' % untarDirName)
        cdAndRun(untarDirName, sh, '%s setup.py install --prefix ~' % sys.executable)
        sh('rm -rf %s' % untarDirName)

    def install(self):
        self._maybeGet()
        self._untarAndInstall()



class CleanUserTest(object):
    def __init__(self, opts):
        self.opt = opts
        
    def clean(self):
        for p in ['bin', 'lib', '_trial_temp', 'certcreate.py', 'server.pem']:
            if osp.exists(p):
                sh('rm -rf %s' % p)

    def test(self):
        o = self.opt
        qplib = xu('~/lib/python2.3/site-packages/quotient')
        assert osp.exists(opj(qplib, 'test', 'testdata', 'ascii', 'alpha.eml'))

        sh("~/bin/trial %s" % o['trialargs'])


    def full(self):
        o = self.opt
        self.clean()

        tardir = o['tardir']
        if osp.exists(tardir):
            if not osp.isdir(tardir):
                sh('rm %s && mkdir %s' % (tardir, tardir))

        for p in o['packages']:
            p.install()

        self.test()



class Options(usage.Options):
    """at least one of {--certcreate, --certfile} must be specified"""

    optParameters = [
        ["product", 'p', 'Quotient',
         "the divmod product we are testing"],
        ["q-ver", None, None,
         "(required) the version of Quotient to download or copy"],
        ['trialargs', None, TRIAL_ARGS,
         "if specified, passes to trial as arguments when run"],
        ['tardir', None, 'tarballs',
         "the location where downloaded tarballs will be placed"]
        ]
    
    optFlags = [
        ["localQball", 'l',
         "copy the quotient tarball from /tmp rather than wgetting it"]
        ]

    def postOptions(self):
        def _mustHave(keys, failMsg=''):
            for k in keys:
                if not self[k]:
                    if not failMsg:
                        raise usage.UsageError("--%s is required!" % k)
                    else:
                        raise usage.UsageError(failMsg)

        def _mustBe(key, okVals):
            if self[key] not in okVals:
                print 'key: %s, self[key]: %s, okVals: %s' % (key, self[key], okVals)
                raise usage.UsageError("--%s must be one of %s" % (self[key], okVals))

        _mustHave(['q-ver'])
        _mustBe('product', ['Quotient', 'Nevow', 'Lupy'])

        Package.pkgdir = self['tardir']
        
        # convenience
        self['qballname'] = "%s-%s.tar.gz" % (self['product'], self['q-ver'])

        reston_url = 'http://osdn.dl.sourceforge.net/sourceforge/'

        lupy = ('Lupy', reston_url + 'lupy/Lupy-0.2.1.tar.gz' )
        spambayes = ('spambayes', reston_url + 'spambayes/spambayes-1.0rc1.tar.gz')
        twisted = ('Twisted', "http://www.twistedmatrix.com/products/get-current.epy?cat=_NoDocs")

        divmod_base = "http://divmod.org/users/slyphon/releases/divmod/"
        nevow = ('Nevow', opj(divmod_base, "Nevow-0.2.0.tar.gz"))

        self['packages'] = [Package(name, url) for name, url in [lupy, spambayes, twisted, nevow]]

        qname, qurl = 'Quotient', None

        if self['localQball']:
            qfile = self['qballname']
            if osp.exists(qfile):
                sh('rm %s' % opj(self['tardir'], qfile))
            sh('cp /tmp/%s %s' % (qfile, self['tardir']))
        else:
            qurl = opj(divmod_base, 'Quotient-%s.tar.gz' % self['q-ver'])

        self['packages'].append(Package(qname, qurl))
            

def main():
    try:
        opts = Options()
        opts.parseOptions()
    except usage.UsageError, ue:
        print "%s: %s (see --help)" % (sys.argv[0], ue)
        sys.exit(2)
#    sys.exit("BYE")

    cut = CleanUserTest(opts)
    cut.full()
        

if __name__ == '__main__':
    main()
