# -*- test-case-name: quotient.test.test_spambayesup -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import re
import os
import time
import rfc822

from nevow import tags
import formless
from formless import webform

from atop.powerup import Powerup
from atop.store import IPipe, Pool, Item, referenced
from atop.tpython import Forgetter
from atop.sched import Schedulable

from twisted.python import log
from twisted.internet import threads, reactor, defer
from twisted.python.components import Interface
from twisted.persisted.styles import Versioned

from quotient.stats import Statoscope
from quotient.mimemessage import MIMEMessage, EmailPowerup, ICleanPoint, IMIMEDelivery, EmailPool
from quotient.proto.smtpout import IOutgoingMail, OutgoingMailup
from quotient import taglib
from quotient.web.root import Base
from quotient.prefup import touchNForward, IPrefs, Prefup
from quotient.application import MailApp, IMailApp

# Hack to get around isinstance(email.Message.Message)
import spambayes.mboxutils as mbu
mbu.get_message = lambda x:x
from spambayes import hammie
from spambayes import __version__ as SBVER
from spambayes.classifier import Bayes
from spambayes.Options import options


class PreferenceGrouping(formless.TypedInterface):
    cleanThresh = formless.Integer(
        label="Clean Threshold",
        description="Above this threshold, messages are marked as clean.",
        required=True)
    spamThresh = formless.Integer(
        label="Spam Threshold",
        description="Below this threshold, messages are marked as spam.",
        required=True)
    incinerateSpam = formless.Boolean(
        label="Incinerate Spam",
        description="Destroy spam after the configured holding period, instead of moving it to trash.")

class ISpamPref(formless.TypedInterface):
    Preferences = PreferenceGrouping

    def resetSpamBayes(self):
        """Forget all SpamBayes training. This operation can not be un-done."""
    resetSpamBayes = formless.autocallable(resetSpamBayes)


class SpambayesupDisplay(Base):
    __implements__ = Base.__implements__, ISpamPref

    # XXX TODO - Get rid of this somehow
    postLocation = property(lambda self: self.original.storeID)

    incinerateSpam = touchNForward("incinerateSpam")

    def spamThreshGet(self):
        return self.original.spamThresh

    def spamThreshSet(self, thresh):
        self.original.touch()
        self.original.spamThresh = thresh
        self.original.setTrainedTime()
    spamThresh = property(spamThreshGet, spamThreshSet)

    def resetSpamBayes(self):
        self.original.setupBayes()

    def cleanThreshGet(self):
        return self.original.cleanThresh

    def cleanThreshSet(self, thresh):
        self.original.touch()
        self.original.cleanThresh = thresh
        self.original.setTrainedTime()
    cleanThresh = property(cleanThreshGet, cleanThreshSet)

    def render_msgsProcessed(self, context, data):
        return self.original.msgsProcessed

    def render_nspam(self, context, data):
        return self.original.guesser.bayes.nspam

    def render_nham(self, context, data):
        return self.original.guesser.bayes.nham

    def render_ntokens(self, context, data):
        return len(self.original.guesser.bayes.wordinfo)

    document = taglib.page(title="SpamBayes Home")[
        ## Todo: Add "editable" flag to formless, so the following can just be described in the TypedInterface
        # tags.p["Messages Processed: ", render_msgsProcessed],
        tags.p["Number of tokens: ", render_ntokens],
        tags.p["Trained Spam: ", render_nspam],
        tags.p["Trained Clean: ", render_nham],
        webform.renderForms()
    ]

    def renderHTTP(self, request):
        getByID = self.original.store.getItemByID
        if request.args.has_key('spam'):
            msgID = request.args['spam'][0]
            msg = getByID(int(msgID))
            self.original.train(msg, 'spam')
        elif request.args.has_key('clean'):
            msgID = request.args['clean'][0]
            msg = getByID(int(msgID))
            self.original.train(msg, 'clean')
        elif request.args.has_key('file'):
            poolToFileIn = request.args['file'][0]
            itemsToFile = request.args['items']
            if poolToFileIn == 'Clean':
                removeFrom = self.original.spamPool
                addTo = self.original.cleanPool
            elif poolToFileIn == 'Spam':
                removeFrom = self.original.cleanPool
                addTo = self.original.spamPool
            else:
                raise RuntimeError, "Misuse of spam powerup"
            def _():
                for itemid in itemsToFile:
                    item = getByID(int(itemid))
                    # An IMAP4 or even another HTTP client may have
                    # done something with this Item since we obtained
                    # its ID.  If that something was removal from this
                    # mailbox, removing it again would be a mistake.
                    # We still want to take whatever action the user
                    # requested though, so the addItem() call is not
                    # protected by this check.
                    if item in removeFrom:
                        removeFrom.removeItem(item)
                    addTo.addItem(item)
            self.original.store.transact(_)
        else:
            return Base.renderHTTP(self, request)
        request.redirect(request.getHeader('referer'))
        return Base.renderHTTP(self, request)


class ISpambayes(Interface):
    """
    @ivar incinerateSpam: If true, indicates that spam should be
    destroyed after the holding period.  Otherwise, spam should
    be moved to the trash pool for manual deletion.
    """

    def train(self, item, pool):
        pass

class Spambayesup(Powerup):
    __implements__ = ISpambayes, Powerup.__implements__, IPipe

    dependencies = [EmailPowerup]
    powerupName = 'SpamBayes'

    # TODO - This should all be moved into Cleanup
    # I wrote this first and then cleanup and did
    # want to change too much at once - AAB

    incinerateSpam = False

    spamThresh = 20
    cleanThresh = 90
    msgsProcessed = 0

    cleanPool = referenced
    spamPool = referenced

    postiniRe = re.compile(r"""\w:\s*\d*\.\d*""")

    persistenceVersion = 5

    def upgradeToVersion5(self):
        self.touch()
        self.spamPool.removePipe(self.fwd())

    def upgradeToVersion4(self):
        self.touch()
        self.spamPool = self.spamRef.getItem()
        self.cleanPool = self.cleanRef.getItem()
        del self.spamRef, self.cleanRef

    def upgradeToVersion3(self):
        self.touch()
        self.setupBayes()
        self.seedBayes()

        sbDir, sbPath = os.path.split(self.sbPath)
        os.unlink(self.sbPath)
        os.rmdir(sbDir)
        del self.sbPath

    def upgradeToVersion2(self):
        self.touch()
        self.activate()
        self.seedBayes()

    def upgradeToVersion1(self):
        self.touch()
        base = self.store.store.filedir
        f = self.sbPath
        self.sbPath = base + '/' + f[f.find('/data/files/')+len('/data/files/'):]

    def setUpPools(self, avatar):
        sbDir = avatar.newDirectory('spambayes')
        self.sbPath = os.path.join(sbDir, 'spambayes.db')
        self.activate()
        avatar.setComponent(ISpambayes, self)
        self.msgsProcessed = 0
        self.setTrainedTime()
        self.pools = {}
        self.setupBayes()
        self.seedBayes()

    def setupBayes(self):
        from spambayes.classifier import Classifier
        def _():
            self.touch()
            self.guesser = hammie.Hammie(Classifier())
            self.seedBayes()
        self.store.transact(_)

    def setTrainedTime(self):
        self.lastTrainedTime = time.time()
        self.touch()

    def train(self, item, pool):
        def _():
            self.touch()
            if pool.lower() == 'spam':
                self.guesser.train(item, True)
            else:
                self.guesser.train(item, False)
            self.setTrainedTime()
        self.store.transact(_)

    def match(self, item):
        pool, prob = self.guessPool(item)
        if pool == 'Spam':
            dest = self.spamPool
        else:
            dest = self.cleanPool
        dest.addItem(item)

    def getPostiniTokens(self, item):
        """
        Postini spam headers look like this:
        X-pstn-levels:     (S:99.9000 R:95.9108 P:91.9078 M:100.0000 C:96.6797 )
        X-pstn-levels:     (S: 0.0901 R:95.9108 P:95.9108 M:99.5542 C:79.5348 )
        """
        levels = item["X-pstn-levels"]
        if levels is None:
            return ''
        else:
            levels = self.postiniRe.findall(levels)
            D = {}
            for lev in levels:
                k, v = lev.split(':')
                v = float(v.strip())
                D[k.upper()] = v
            # spam_level is counter-intuitive: lower means more spammy
            # For now, we only care about the 'S' value, which is the overall
            # spam score
            spam_level = D['S']
            for i in range(1,6):
                if spam_level <= i:
                    return'Postini_spam_level_very_low_' + str(i)
                    return tok
            if spam_level <= 30:
                return 'Postini_spam_level_low'
            if spam_level <= 60:
                return 'Postini_spam_level_medium'
            if spam_level <= 94:
                return 'Postini_spam_level_high'
            for i in range(95,101):
                if spam_level <= i:
                    return 'Postini_spam_level_very_high_' + str(i)

    def seedBayes(self):
        self.guesser.bayes.learn(['Postini_spam_level_very_low_1',
                                  'Postini_spam_level_very_low_2'], True)
        self.guesser.bayes.learn(['Postini_spam_level_very_low_100'], False)
        def _():
            self.setTrainedTime()
            self.touch()
        self.store.transact(_)

    def guessPool(self, item):
#       XXX: ROAR! BROKEN-NESS!!
#        from quotient import sysup
#        cp = sysup.IControlPanel(self.store.store)
#        perm = sysup.IControlPanel.SPAMBAYES_TRAINING
#        if not cp.actionAllowed(perm):
#            return ('Clean', 0.5)

        finfo = getattr(item, 'filingInfo', None)
        if finfo:
            pool, prob, lTime = finfo
            if pool is None:
                # Message filed via a rule
                # return (None, prob)
                pass
            else:
                if not self.lastTrainedTime > lTime:
                    # cache hit
                    return (pool, prob)

        toks = hammie.tokenize(item)
        # Could iterate tokens in a reactor before calling
        # spamprob
        toks = list(toks)
        pTok = self.getPostiniTokens(item)
        if pTok != '':
            toks.append(pTok)
        prob = self.guesser.bayes.spamprob(toks, False)

        prob = 1.0 - prob
        if prob > self.cleanThresh/100.0:
            pool = 'Clean'
        elif prob < self.spamThresh/100.0:
            pool = 'Spam'
        else:
            pool = 'Unknown'
        item.filingInfo = (pool, prob, time.time())
        item.touch()
        return (pool, prob)

    def setCleanAndSpam(self, clean, spam):
        """
        This needs to be generalized for Reverend, which can
        handle multiple aribtrary pools
        """
        self.cleanPool = clean
        self.spamPool = spam

class IBasicWhitelist(formless.TypedInterface):
    defaultFrom = formless.String(label="Add", description="")

class WhitelistDisplay(Base):
    __implements__ = IBasicWhitelist, Base.__implements__
    def render_whitelist(self, context, data):
        addys = self.original.whitelist.list.keys()
        addys.sort()
        disp = []
        for a in addys:
            disp.append(tags.p["%s" % a])
        if disp:
            return tags.p["Addresses in your whitelist:"], disp
        return tags.p["There are no addresses in your whitelist."]

    document = taglib.page(title="Whitelist Home")[
        tags.p["Add an address to your whitelist:"],
        tags.form(id='addAddress',action="")[
            tags.input(type="text", name="add"),
            tags.input(type="submit")
        ],
        render_whitelist
    ]

    def renderHTTP(self, request):
        if request.args.has_key('add'):
            addy = request.args['add'][0]
            def _():
                self.original.addToWhitelist(addy)
            self.original.store.transact(_)
            request.redirect(request.getHeader('referer'))
            return Base.renderHTTP(self, request)
        return Base.renderHTTP(self, request)

class IRuler(Interface):
    def match(self, item):
        """Return True or False"""

class IRule(Interface):
    def match(self, item):
        """Return True or False"""

class Whitelist(object):
    __implements__ = IRule

    def __init__(self):
        self.list = {}

    def match(self, item):
        name, address = item.nameAndAddress()
        address = address.lower()
        if address in self.list:
            return True
        return False

    def addToList(self, address):
        self.list[address.lower()] = True

class Ruler(Powerup):
    __implements__ = IRuler, IPipe, Powerup.__implements__
    dependencies = [OutgoingMailup, Prefup]
    powerupName = 'Whitelist'

    def setUpPools(self, avatar):
        avatar.setComponent(IRuler, self)
        self.rules = {}
        self.whitelist = Whitelist()
        outmail = IOutgoingMail(avatar)
        outmail.sentRef.getItem().addPipe(self.fwd(IPipe))

    def match(self, item):
        for name, rule in self.rules.items():
            if rule.match(item):
                item.filingInfo = (None, name, None)
                item.touch()
                return True
        return False

    def addRule(self, ruleName, rule):
        self.touch()
        self.rules[ruleName] = rule

    def addToWhitelist(self, address):
        self.touch()
        self.whitelist.addToList(address)

    def itemAdded(self, item):
        """
        Automatically add folks that you sende
        a message to, to your whitelist.
        """
        if IPrefs(self.store).whitelistOnSend:
            addys = ''
            for field in ['to','cc','bcc']:
                addy = item[field]
                if addy:
                    addys += addy + ','
            addresses = [x[1] for x in rfc822.AddressList(addys)]
            # assert addresses, 'There must be a to, cc or bcc header'
            for wl in addresses:
                self.addToWhitelist(wl)

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, changeData):
        pass

class PostiniVirusRule:
    def match(self, item):
        """
        Postini virus headers look like this:
        X-pstnvirus: W32/Mydoom.a@MM
        """
        virusInfo = item["X-pstnvirus"]
        if virusInfo is not None:
            ruleInfo = "Virus! " + virusInfo
            item.filingInfo = ('Virus', ruleInfo, None)
            item.touch()
            return True
        else:
            return False

class IClean(Interface):
    pass

class Cleanup(Powerup):
    __implements__ = (IClean, IPipe, Powerup.__implements__)
    dependencies = [Spambayesup, Ruler, MailApp]
    powerupName = 'Cleaner'

    spamPool = referenced
    cleanPool = referenced
    virusPool = referenced

    persistenceVersion = 4

    def upgradeToVersion4(self):
        self.touch()
        self.cleanPool.setIndexKey('name', 'Inbox')

    def upgradeToVersion3(self):
        self.touch()

        self.spamPool = self.spamRef.getItem()
        self.cleanPool = self.cleanRef.getItem()
        self.virusPool = self.virusRef.getItem()
        del self.spamRef, self.cleanRef, self.virusRef

    def upgradeToVersion2(self):
        self.touch()

        from atop.powerup import IPowerStation
        from quotient.prefup import IPrefs

        ignore = IPrefs(self.store, default=None)
        app = IPowerStation(self.store).installPowerup(MailApp)

        app.appRoot.addItem(self.cleanRef.getItem())
        app.appRoot.addItem(self.spamRef.getItem())
        app.appRoot.addItem(self.virusRef.getItem())
        self.store.getRootPool().removeItem(self.cleanRef.getItem())
        self.store.getRootPool().removeItem(self.spamRef.getItem())
        self.store.getRootPool().removeItem(self.virusRef.getItem())

    def upgradeToVersion1(self):
        self.touch()
        self.virusRule = PostiniVirusRule()
        virus = EmailPool(self.store, 'Virus')
        self.virusRef = virus.referenceTo()
        self.store.getRootPool().addItem(virus)

    def activate(self):
        pass

    def setUpPools(self, avatar):
        arrRef = IMIMEDelivery(avatar).getArrivalRef()
        arrRef.getItem().addPipe(self.fwd(IPipe))
        avatar.setComponent(IClean, self)
        self.rulerRef = IRuler(avatar).referenceTo()
        app = IMailApp(avatar)
        sb = ISpambayes(avatar)
        self.sbRef = sb.referenceTo()
        spam = EmailPool(avatar, 'Spam')
        virus = EmailPool(avatar, 'Virus')
        clean = EmailPool(avatar, 'Inbox') # Was Clean!!!
        self.spamPool = spam
        self.cleanPool = clean
        self.virusPool = virus
        self.virusRule = PostiniVirusRule()
        app.appRoot.addItem(clean)
        app.appRoot.addItem(spam)
        app.appRoot.addItem(virus)
        sb.setCleanAndSpam(clean, spam)
        # Move the clean point
        cp = ICleanPoint(avatar)
        cp.getItem().setPoint(clean)

    def itemAdded(self, item):
        if not isinstance(item, MIMEMessage):
            return
        # First check for viruses
        if self.virusRule.match(item):
            self.virusPool.addItem(item)
        elif self.rulerRef.getItem().match(item):
            # Check the whitelist
            self.cleanPool.addItem(item)
        else:
            # Pass it on to SpamBayes
            self.sbRef.getItem().match(item)

    def itemRemoved(self, item):
        pass

    def itemChanged(self, item, changeData):
        pass

class HeaderRule(object):
    __implements__ = (IRule,)

    def __init__(self, headerName, headerValue):
        self.name = headerName
        self.value = headerValue

    def match(self, item):
        return self.op(item[self.name], self.value)

class HeaderEquals(HeaderRule):
    def op(self, orig, spec):
        return orig.lower() == spec.lower()

class HeaderDoesntEqual(HeaderRule):
    def op(self, orig, spec):
        return orig.lower() != spec.lower()

class HeaderContains(HeaderRule):
    def op(self, orig, spec):
        return orig.lower().find(spec.lower()) != -1

class HeaderDoesntContain(HeaderRule):
    def op(self, orig, spec):
        return orig.lower().find(spec.lower()) == -1

class HeaderMatchesExpression(HeaderRule):
    _cache = None
    def op(self, orig, spec):
        if self._cache is None:
            self._cache = re.compile(spec)
        return self._cache.match(orig) is not None

class HeaderDoesntMatchExpression(HeaderRule):
    _cache = None
    def op(self, orig, spec):
        if self._cache is None:
            self._cache = re.compile(spec)
        return self._cache.match(orig) is not None

def HeaderRuleDisplay(context, data):
    return tags.form(action=str(data.storeID))[
        "Header:",
        tags.input(type="text", name="headerName"),
        "Operation:",
        tags.input(type="text", name="operation"),
        "Value:",
        tags.input(type="text", name="headerValue"),
        ]

class Deduplicator(Item, Schedulable):
    clean = referenced
    spam = referenced
    dupes = referenced

    _headers = ['from', 'to', 'subject', 'date', 'message-id']
    started = False

    def __init__(self, st, clean, spam, dupes):
        Item.__init__(self, st)
        self.allByTime = {}
        self.clean = clean
        self.spam = spam
        self.dupes = dupes

    def __repr__(self):
        return '<Deduplicator (%s)>' % (self.store,)

    def __call__(self, pool, item):
        self.touch()
        when = item.dateReceived
        hdrs = [item[h] for h in self._headers]
        for p in (self.clean, self.spam):
            msgs = p.queryIndex('received', startKey=when, endKey=when)
            for m in msgs:
                if m is not item:
                    if [m[h] for h in self._headers] == hdrs:
                        self.removeDuplicate(item, m)

    def removeDuplicate(self, keep, discard):
        for p in discard.iterpools():
            keep.changeFlags(p, discard.getIndexKey(p, 'flags', 0))
        keep.changeFlags(None, discard.getIndexKey(None, 'flags', 0))
        pop3uid = discard.getIndexKey(None, 'pop3uid', None)
        if pop3uid is not None:
            keep.setIndexKey('pop3uid', pop3uid)
        for p in discard.iterpools():
            if discard in p:
                p.removeItem(discard)
        self.dupes.addItem(discard)

def deduplicate(avatar):
    from atop import upgrade
    app = IMailApp(avatar, default=None)
    sb = ISpambayes(avatar, default=None)
    md = IMIMEDelivery(avatar, default=None)

    if app is not None and sb is not None and md is not None:
        dupes = app.appRoot.getIndexItem('name', 'Duplicate Messages', None)
        if dupes is None:
            dupes = Pool(avatar, 'Duplicate Messages')
            app.appRoot.addItem(dupes)
        arrival = md.getArrivalRef().getItem()
        if sb.cleanPool is not None and sb.spamPool is not None:
            deduper = Deduplicator(avatar, sb.cleanPool, sb.spamPool, dupes)
            u = upgrade.ItemUpgradeFactoryUpgradeQueue(avatar, arrival, deduper)
            return u.start()
    log.msg("Skipping upgrade for %s" % (avatar.name,))
