# -*- test-case-name: quotient.test.test_extract -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import re

from atop import store
from atop import tpython
from atop.powerup import Powerup

from nevow import tags

from quotient import cleanup
from quotient import provisioner
from quotient import mimemessage
from quotient import taglib
from quotient.web.root import Base
from quotient.note import Note
from quotient.smarts.todo import INotebook
from quotient import searchup

from twisted.python.components import Interface

class ICommand(Interface):
    def commandRaw(self, request, cmdStr):
        pass

    def command(self, request, cmd, arg):
        pass

class Commander(Powerup):
    dependencies = []
    invisible = True
    totalExtracted = 0
    powerupName = "Commander"
    _implements__ = ICommand,

    poolOfCommands = store.referenced

    def setUpPools(self, avatar, name=None):
        avatar.setComponent(ICommand, self)
        commandsPool = store.Pool(avatar, "Commands")
        self.poolOfCommands = commandsPool
        avatar.getRootPool().addItem(commandsPool)

    def commandsPool(self):
        return self.poolOfCommands

    def commandRaw(self, request, cmdStr, _charset_):
        order = Note(self.store, cmdStr)
        self.commandsPool().addItem(order)
        request.redirect('/me/%s?query=%s&_charset_=%s' % (searchup.ISearch(self.store).storeID, cmdStr, _charset_))
        return ''

    def command(self, request, cmd, arg):
        pass

class CommanderDisplay(Base):
    document = taglib.page(title="Commander")[
        tags.p["Hello"],
    ]
