# -*- test-case-name: "quotient.test.test_webui" -*-
import urllib, re

import os.path as osp
from os.path import join as opj

from twisted.web.domhelpers import findElementsWithAttribute as findewa
from twisted.python import log, urlpath, components, failure
from twisted.web.domhelpers import findNodes
from twisted.web.domhelpers import findNodesNamed as findnn
from twisted.web.domhelpers import gatherTextNodes as gather
from twisted.web.microdom import parseXMLString
from twisted.web import microdom, resource, domhelpers
from twisted.protocols import sux
from twisted.trial.unittest import FailTest

from quotient.functional import httpclient

_client_debug = False

def debugObserver(events):
    if _client_debug and 'client' in events:
        print ' '.join(events['client'])
log.addObserver(debugObserver)

def _dbgmsg(msg):
    log.msg(client=msg)


class Bogus(object):
    """does nothing, and succeeds"""
    def __init__(self, *a, **kw): pass
    def __call__(self, *a): return self
    def __str__(self): return "Bogus() object"
    def __repr__(self): return "<Bogus() object @%s>" % hex(id(self))
    def __getattr__(self, a): 
        self._warn("__getattr__", a)
        return self
    def __setattr__(self, *a): 
        self._warn("__setattr__", a)
        return self
    def __delattr__(self, a): 
        self._warn("__delattr__", a)
        return self
    def _warn(self, method, args=None, kwargs=None):
        log.msg(client="WARNING!!!: Bogus.%s called with args: %s, kwargs: %s\n THIS IS MOST LIKELY AN ERROR!!!" % (method, str(args), str(kwargs)))

try:
    page
except NameError:
    page = Bogus()


class FormParsingError(Exception):
    pass

class FormElementParsingError(FormParsingError):
    pass

class IFormElement(components.Interface):
    """Represents an html form element in an IForm
    @cvar type: a string describing the type of this element ('radio', 'textbox', etc.)
    @type type: string
    @cvar id: the id given to this form element (ex. 'username')
    @type id: string
    @cvar value: the value this element holds
    @type value: string
    """
    pass


class FormElement:
    __implements__ = IFormElement,
    id    = None
    node  = None
    type  = None
    name  = None
    value = None

    def __init__(self, node, form=None):
        # TODO: Get rid of this nonsense and override __getattr__
        self.node = node
        self.id = node.getAttribute('id')
        self.type = node.getAttribute('type')
        self.name = node.getAttribute('name')
        self.value = node.getAttribute('value')


class IForm(components.Interface):
    """An abstraction of a web form
    @ivar id: the id of this form
    @type id: string
    @ivar node: the node representing this form
    @type node: twisted.web.microdom.Node
    @ivar page: the page that contains this form (this form's parent object)
    @type page: an IPage
    @ivar action: this forms action attribute's value
    @type action: string
    """
    def __init__(self, node, page=None):
        """
        @param node: the node representing this form
        @type node: twisted.web.microdom.Node
        """
        pass

    def submit(self, args=None): # OR could the form elements be filled in then submit() with no args?
        """submits this form to the server with the appropriate argument
        string as defined by kw
        @param args: key, value pairs to be submitted
        @type args: dict
        """
        pass


class Form(object):
    __implements__ = IForm,

    id, node, page, action, _element = None, None, None, None, None

    def __init__(self, node, page=None):
        self.node, self.page = node, page
        self.id = node.getAttribute('id')
        self.action = a = node.getAttribute('action')
        if not a:
            self.action = self.page.url
        
    def __str__(self):
        from cStringIO import StringIO
        sio = StringIO()
        for n in ['id','node','page','action']:
            sio.write("%s: %s\n" % (n, getattr(self, n, None)))
        return sio.getvalue()

    def _loadElements(self):
        if self._element is None:
            self._element = dict([(f.id, f) for f in [FormElement(i, self) for i in findnn(self.node, 'input')]])
        return self._element
    elements = property(_loadElements)

    def getElement(self, id):
        return self.elements[id]

    def submit(self, args=None):
        ("Form.submit()")
        if self.page is None:
            raise RuntimeError, "Can't follow link, no page object"
        if args is None:
            args = dict([(e.name, e.value) for e in self.elements.iteritems()])
        action = self.action
        if not action:
            action = self.page.url
        return self.page.client.get(action, args)


class ILink(components.Interface):
    """An abstraction of a web link
    @ivar id: a unique id
    @type id: string
    @ivar name: the display name of this link
    @type name: string
    @ivar url: the url to which this link points
    @type url: string
    @ivar node: this link's microdom node
    @type node: twisted.web.microdom.Element
    """
    pass

# XXX: this is broken and probably unnecessary:

class LinkNode(object, components.Adapter):
    """an adapter for ILink, allows for Links that don't
    have 'id' attributes"""

    __implements__ = ILink,

    def __init__(self, original):
        node = self.node = self.original = original
        self.link = Link(node, failUnlessId=False)

    def _getPage(self):
        return self.link.page

    def _setPage(self, page):
        self.link.page = page

    page = property(_getPage, _setPage)

    def _getUrl(self):
        return self.link.url

    def _getId(self):
        return self.link.id

    def _getName(self):
        return self.link.name

    url, id, name = property(_getUrl), property(_getId), property(_getName)

    def follow(self):
        self.link.follow()



class IdNotFoundError(failure.DefaultException):
    pass

class Link(object):
    __implements__ = ILink,

    id, name, url, node = None, None, None, None

    def __init__(self, node, page=None, failUnlessId=True):
        self.node = node
        self.page = page
        nid = node.getAttribute('id')
        if not nid and failUnlessId:
            raise IdNotFoundError()
        self.id = nid
        self.url = node.getAttribute('href')
        self.name = gather(node)

    def __str__(self):
        return "%s: %s" % (self.name, self.url)

    def follow(self):
        # log.msg('Link follow()')
        if self.page is None:
            raise RuntimeError, "Can't follow link, no page object"
        return self.page.client.get(self.url)


class FunctionalError(KeyError):
    def __init__(self, id, page):
        self.id = id
        self.page = page

    def __str__(self):
        from os.path import abspath
        return """\n\n%s with id=%s\n was not found in page: %s\n
which was saved to: %s
the urlrail can be found at: %s
and vieweed in your  web browser at file://%s
""" % (self._name, self.id, page.url,
       urlTrail.filename,
       urlTrail.path,
       abspath(urlTrail.path))


class NoSuchLinkError(FunctionalError):
    _name = 'link'

class NoSuchFormError(FunctionalError):
    _name = 'form'
    def __str__(self):
        return '\n\nform with id=%s\n\n not found in page: %s\n\n' % (self.id, self.page.url)


class IPage(components.Interface):
    """an abstraction of a web page
    @ivar dom: a string containing the dom this IPage represents
    """
    def getForm(self, id):
        """returns the IForm object named name on this page
        @param name: the name of the form to return
        @return: an IForm object
        @rtype: IForm
        @raise NoSuchFormError: if no form 'name' could be found
        """
        pass

    def getLink(self, id):
        """returns an ILink object from this page
        @param id: the value of the id attribute of the link
        @return: an ILink object
        @raise NoSuchLinkError: if the link 'name' could not be found
        """
        pass

from twisted.protocols.sux import ParseError

class Page(object):
    __implements__ = IPage,

    _linkNameIdx = None
    domstr = ''
    _dom, _forms, _links, _allLinks, _url, _title, encoding = None, None, None, None, None, None, None
    _rTime = -1.0

    def __init__(self, domstr, client=None):
        """construct
        @param domstr: the dom string that represents this page of information
        @type domstr: a unicode object
        @param client: the client that generated this page (a.k.a. 'parent')
        @type client: a L{quotient.web.functional.Client} instance
        """
        self.domstr, self.client = domstr, client

    def _getdom(self):
        if not self._dom:
            try:
                self._dom = parseXMLString(self.domstr) #.encode(ENCODING))
            except ParseError, e:
                log.err(e)
                log.msg(self.domstr)
                raise
        return self._dom
    dom = property(_getdom)

    def _loadforms(self):
        if not self._forms:
            self._forms = dict([(f.id, f) for f in
                    [Form(n, self) for n in findnn(self.dom, 'form')
                     if n.getAttribute('value') != 'search']]) # XXX: For now skip "search" form
        return self._forms
    forms = property(_loadforms)

    def _loadlinks(self):
        if not self._links:
            linklist = []
            for node in findnn(self.dom,'a'):
                try:
                    linklist.append(Link(node, self))
                except IdNotFoundError, e:
                     continue
            self._links = dict([(L.id, L) for L in linklist])
        return self._links
    links = property(_loadlinks)

    def _loadAllLinks(self):
        if not self._allLinks:
            self._allLinks = []
            for l in [ILink(node) for node in findnn(self.dom, 'a')]:
                l.page = self
                self._linkNameIdx[l.name] = l
                self._allLinks.append(l)
        return self._allLinks
    allLinks = property(_loadAllLinks)

    def _loadTitle(self):
        if not self._title:
            self._title = gather(self.dom.getElementsByTagName('title')[0])
        return self._title
    title = property(_loadTitle)

    def _getLinkNameIdx(self):
        if not self._linkNameIdx:
            self._loadAllLinks()
        return self._linkNameIdx
    linkNameIdx = property(_getLinkNameIdx)

    def findLinkNamed(self, name):
        """returns link with name on this page, returns None if not found"""
        try:
            return self.linkNameIdx[name]
        except KeyError:
            return None

    def _getRenderTime(self):
        if self._rTime == -1.0:

            try:
                s = gather(findewa(self.dom, 'id', 'timer')[0])
            except IndexError, e:
                s = None            # XXX; this is dumb

            if s:
                self._rTime = float(s.replace(' ms', ''))
            else:
                self._rTime = None

        return self._rTime
    renderTime = property(_getRenderTime)

    def getForm(self, lid):
        if lid not in self.forms:
            raise NoSuchFormError(lid, self)
        return self.forms[lid]

    def getLink(self, lid):
        if lid not in self.links:
            raise NoSuchLinkError(lid, self)
        return self.links[lid]

    def hasLink(self, lid):
        return lid in self.links


def encode_multipart_formdata(aDict):
    L = []
    b = "---------96329786497863496029361092386519203685-------"
    CRLF = '\r\n'

    for (key, value) in aDict.items():
        L.append('--%s' % b)
        L.append('content-disposition: form-data; name="%s"' % key)
        L.append(value)
    L.append('--%s--' % b)

    body = CRLF.join(L)
    
    headers = {'content-type': 'multipart/form-data; boundary=%s' % (b),
               'content-length': str(len(body))} 
   
    return headers, body
                

try:
    port
except NameError:
    port = 80

try:
    numRequests
except NameError:
    numRequests = 0

try:
    domain
except NameError:
    domain = 'localhost'



try:
    urlTrail
except NameError:
    urlTrail = None #Bogus()

class ClientError(Exception):
    pass

class BogusDomError(ClientError):
    pass

class Client:
    """retrieves a page (?) and provides a convenient API for posting to its forms and
    inspecting its content
    """
    __debug = False
    _opener = httpclient.opener()
    url, data = None, None
    path = ''

    def __init__(self):
        log.theLogPublisher.observers.insert(0, self._debugObserver)

    def _debugObserver(self, events):
        if 'Client' in events:
            if self.__debug:
                print ' '.join(['Client:\t'] + list(events['Client']))

    def _dbgmsg(self, msg):
        log.msg(Client=msg, message=None)

    def get(self, path, args={}):
        """
        @return: a deferred that will deliver the dom string
        """
        ## @rtype: a unicode object

        if self.url is None:
            self.url = url = "http://%s:%d%s" % (domain, port, path)
        else:
            self.url = url = str(urlpath.URLPath.fromString(self.url).click(path))

        def doRequest(request):
            self._dbgmsg('doRequest')
            d = self._opener.open(request
                                 ).addCallback(httpclient.read
                                 ).addCallback(mostBasicDomTest
                                 ).addCallback(createPage
                                 ).addErrback(dealWithError)
            return d

        def mostBasicDomTest(result):
            self._dbgmsg('mostBasicDomTest')
            if result[0] != '<':
                raise BogusDomError(result)
            return result

        def createPage(result):
            self._dbgmsg('createPage')
            global page
            page = Page(result, self)
            page.url = self.url + self.data

            # the following makes use of a Bogus object, 
            # if urltrail hasn't been started, or has been stopped
            # any actions will produce no results and will be logged
            urlTrail.savePage(page.url, page.domstr) 
            return page

        def dealWithError(error):
            self._dbgmsg('dealWithError')
            log.msg('trapped exception retreiving url %s' % self.url)

            error.trap(Exception)
            log.msg(error.getErrorMessage())
            if "Connection was closed cleanly" in error.getTraceback():
                return None
            print 'dealWithError: %s' % error
            raise error
 
        data = ''
        headers = {}
        if args:
            if 'multipart' in args and args['multipart']:
                del args['multipart']
                headers, data = encode_multipart_formdata(args)
            else:
                urlQuery = '?'
                for argKey, argValues in args.items():
                    if isinstance(argValues, str):
                        argValues = [argValues]
                    for val in argValues:
                        urlQuery += '%s=%s&' % (urllib.quote(argKey), urllib.quote(val, safe=''))
                url = url + urlQuery[:-1]

        self.data = data
        req = httpclient.Request(url, data, headers)
        return doRequest(req)
  

## WE ARE BORG! RESISTANCE IS FUTILE! ##
# okay, this isn't _exactly_ the Borg Non-Pattern,
# but how about a bit of artistic license!?

try:
    _theClient
except NameError:
    _theClient = Client()


def get(path, args={}):
    global numRequests
    numRequests += 1
    return _theClient.get(path, args)


# XXX: Need to figure out a way to handle the urltrail file cleanly, without
# opening and closing it with each write

class UrlTrail(object):
    _curIdNum = 0
    reCleanTitle = re.compile(r'[_ /\\.,-]')

    def __init__(self, tmpDir, trailName):
        # TODO:
        # make the urltrail file at the same level in the directory heirarchy as the
        # test.log file, so that it's refreshable in a browser
        
        self._tmpDir, self._trailName = tmpDir, trailName
        self._fd = file(self.path, 'w')
        self._fd.write("""\
<html><head><title>urltrail for test run</title></head>
<body><ol>
""")
        log.debug('urltrail path: file://%s' % self.path)

    def _getUrlTrailPath(self):
        return opj(self._tmpDir, self._trailName)
    path = property(_getUrlTrailPath, None, None, "the path to the urlTrail file")
    
    def write(self, astr):
        """for writing information to the urltrail"""
        self._fd.write(astr)
        self._fd.flush()

    def _getFilename(self, title=page.title):
        cleanTitle = lambda t: self.reCleanTitle.sub(lambda _: '_', t)
        return '%02d_%s.html' % (self._curIdNum, cleanTitle(page.title))

    # XXX: rename this
    filename = property(_getFilename, None, None, 'the name that I will save the page to')

    def savePage(self, url, domstr):
        """I save a retrieved dom string from url to the next sequential filename and
        make a link in the urlTrail file"""
        self._curIdNum += 1
        fname = self.filename
        f = file(opj(self._tmpDir, fname), 'w')
        f.write(''.join(['<!--%s-->\n' % url, domstr]))
        f.flush(); f.close()
        self.write("""<li><a href=%s>%s: %s</a></li>""" % (fname, fname, url))

    def close(self):
        if not self._fd.closed:
            self._fd.close()
            return
        raise RuntimeError, 'already closed UrlTrail!'


def startUrlTrail(path, name):
    """I start the client logging retrieved pages to the urlTrail
    note: if you called stopUrlTrail before, you will overwrite the existing
    urltrail file by calling this a second time. this may change in the future"""
    global urlTrail
    urlTrail = UrlTrail(path, name)

    
def stopUrlTrail():
    """I close the urltrail file and stop the client from logging further pages"""
    global urlTrail
    urlTrail.close()
    urlTrail = Bogus()


