# -*- test-case-name: quotient.test.test_note -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import time

from atop.store import Item, Pool, registerIndex, referenced, indexed
from twisted.python.components import Interface
import formless

class INote(Interface):
    """
    An Item that has:

    @ivar name: which is a brief, indexable unique identifier for use on the
    command line and within other areas of the system.
    
    @ivar summary: which is a slightly less brief, pretty human-readable string
    which describes the object.
    
    @ivar text: which is a potentially largeish (~32k chars) unicode text/plain
    'body' of the object.
    
    @ivar creator: which is a reference to a Person which is responsible for
    this item.  (the sender, for example)
    
    @ivar creationDate: which is the date that the item entered the system.
    """

class IEditableNote(formless.TypedInterface):
    text = formless.Text()

CREATION_DATE = registerIndex(int, "creation_date")
CREATOR = registerIndex(Item, "creator")

class Note(Item):
    __implements__ = INote, IEditableNote

    creator = indexed(CREATOR)
    creationDate = indexed(CREATION_DATE)

    def __init__(self, store,
                 name='No name',
                 summary='No summary',
                 text='No text',
                 creator=None,
                 creationDate=None):
        Item.__init__(self, store, name=name)
        self.summary = summary
        self.text = text
        self.creator = creator
        self.creationDate = creationDate or time.time()

    def __repr__(self):
        return '<Note: %s>' % self.text
