from twisted.internet import reactor, stdio
from twisted.spread import pb
from twisted.protocols.basic import LineReceiver
from twisted.python import usage
from twisted.cred.credentials import UsernamePassword

class PortholeClient(LineReceiver):

    delimiter = '\n'
    
    def __init__(self, client):
        self.client = client
        self.client.callRemote("getRootPool").addCallback(self._cbStartup)
        self.id = 0
    def _cbStartup(self, root):
        self.current = root
        self.printAndPrompt()
        
    def lineReceived(self, line):
        args = line.strip().split(" ", 1)
        cmd = args[0].lower()
        if cmd:
            if hasattr(self, "cmd_%s"%cmd):
                getattr(self, "cmd_%s"%(cmd))(*args[1:]).addCallback(self.printAndPrompt).addErrback(self.exceptAndPrompt)
            else:
                self.transport.write("Unknown command '%s'\n"%(cmd))
        else:
            self.printAndPrompt()
            
    def cmd_scrutinize(self):
        return self.current.callRemote("scrutinize")

    def _selectItem(self, item, id):
        self.current = item
        self.id = id

    def cmd_select(self, id):
        id = int(id)
        return self.client.callRemote("select", id).addCallback(self._selectItem, id)
    
    def cmd_su(self, username):
        return self.client.callRemote("selectUser", username).addCallback(lambda x :self._selectItem(*x))

        
    def cmd_list(self):
        return self.current.callRemote("list").addCallback(
            lambda x: '\n'.join(
            ["%s: %s <%s>" % (id, name, classname)
             for (id, obj, name, classname) in x]))

    def cmd_eval(self, code):
        return self.current.callRemote("eval", code)

    def exceptAndPrompt(self, failure):
        if failure.type.endswith("PortholeError"):
            self.transport.write(failure.value+'\n')
        else:
            failure.printTraceback(self.transport)
        self.transport.write("%s>> " % self.id)
        
    def printAndPrompt(self, results=None):
        if results:
            self.transport.write(results+"\n")
        self.transport.write("%s>> " % self.id)

class PortholeOptions(usage.Options):

    optParameters = [
        ['host', 'h', 'localhost', "Host running porthole"],
        ['port', 'p', '8787', "Port for porthole"],
        ['username', 'u', '', "the Quotient user to login as"],
        ['passwd', 'w', '', "password for the Quotient user"],
        ]

    

def err(x):
    print x

def main():
    config = PortholeOptions()
    config.parseOptions()
    pbc = pb.PBClientFactory()
    reactor.connectTCP(config['host'], int(config['port']), pbc)
    pbc.login(UsernamePassword(config['username'],
                              config['passwd'])).addCallback(_cbLogin).addErrback(err)
    reactor.run()

def _cbLogin(ava):

    UI= PortholeClient(ava)
    stdio.StandardIO(UI)

    
