# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from atop import credup
from atop import powerup
from atop.store import Item, Pool, referenced

from twisted.python import log
from twisted.cred import checkers
from twisted.python import components
from twisted.application import internet


import formless
from formless import webform
from nevow import tags
from nevow.url import here
from nevow import inevow
from quotient import taglib
from quotient.web.root import Base


class INetworkService(formless.TypedInterface):
    class IEverything(formless.TypedInterface):
        serviceName = formless.String()
        interface = formless.String()
        port = formless.Integer()

    def start(self):
        pass
    start = formless.autocallable(start)

    def stop(self):
        pass
    stop = formless.autocallable(stop)


def wahwahwah(name):
    def get(self):
        return self.pKwArgs[name]
    def set(self, value):
        self.pKwArgs[name] = value
        if self.running:
            self.stop()
            self.start()
    return property(get, set)


def setNTouch(name):
    def get(self):
        return getattr(self, name)
    def set(self, value):
        self.touch()
        setattr(self, name, value)
    return property(get, set)


class NetworkServer(Item):
    __implements__ = INetworkService

    serviceName = setNTouch('name')
    interface = wahwahwah('interface')
    port = wahwahwah('port')

    persistenceForgets = ['port']

    lPort = None
    running = False

    def __init__(self, store, name, ServiceType, kwargs):
        Item.__init__(self, store, name)
        self.pType = ServiceType
        self.pKwArgs = kwargs

    def start(self):
        if not self.running:
            self.lPort = self.pType(**self.pKwArgs)
            self.lPort.startService()
            self.running = True

    def stop(self):
        if self.running:
            self.lPort.stopService()
            del self.lPort
            self.running = False

class IServer(components.Interface):
    def addService(self, name, type, kwargs):
        pass

class Serveup(powerup.Storeup):
    net = referenced

    powerupName = "Port Configuration"

    def setUpPools(self, store):
        store.setComponent(IServer, self)
        store.service.notifyOnStartup(self)
        self.net = Pool(store, "Network Servers")

    def privilegedStartup(self):
        def _():
            for s in self.net.queryIndex():
                s.start()
        self.store.transact(_)

    def startup(self):
        pass

    def addService(self, name, type, kwargs):
        def _():
            s = NetworkServer(self.store, name, type, kwargs)
            self.net.addItem(s)
        self.store.transact(_)

    def _store(self):
        return self.store

    def _makePortal(self):
        # XXX This logic is absolutely misplaced.
        from twisted.cred import portal
        from twisted.cred import checkers
        s = self._store()
        r = credup.FwdRealm(portal.IRealm(s).referenceTo())
        c = credup.FwdChecker(checkers.ICredentialsChecker(s).referenceTo())
        p = portal.Portal(r)
        p.registerChecker(c)
        p.registerChecker(checkers.AllowAnonymousAccess())
        return p

    def addHTTPServer(self, interface, port):
        from twisted.web.woven import guard
        from twisted.application import internet
        from quotient.web import utility
        from quotient.web import root
        upw = guard.UsernamePasswordWrapper(
            self._makePortal(),
            callback=utility.loginCallback,
            errback=utility.badLoginCallback
            )
        site = root.AtopSite(guard.SessionWrapper(upw))
        args = dict(port=port, interface=interface, factory=site)
        self.addService('HTTP Server', internet.TCPServer, args)

    def addSMTPServer(self, interface, port):
        from twisted.application import internet
        from quotient.tap2 import AUTHORIZERS
        from quotient.proto.smtpin import ESMTPInFactory
        s = self._store()
        smtpd = ESMTPInFactory(s, self._makePortal(), AUTHORIZERS, None)
        args = dict(port=port, interface=interface, factory=smtpd)
        self.addService('SMTP Server', internet.TCPServer, args)

    def addIMAPServer(self, interface, port):
        from quotient.proto.imapup import IMAP4Factory
        from quotient.tap2 import AUTHORIZERS
        imapd = IMAP4Factory(self._makePortal(), AUTHORIZERS, None)
        args = dict(port=port, interface=interface, factory=imapd)
        self.addService('IMAP Server', internet.TCPServer, args)

    def addPOPServer(self, interface, port):
        from quotient.proto.popup import POP3Factory
        popd = POP3Factory(self._makePortal(), None)
        args = dict(port=port, interface=interface, factory=popd)
        self.addService('IMAP Server', internet.TCPServer, args)

    def addPBServer(self, interface, port):
        from quotient.proto.pbup import PBServerFactory
        pbd = PBServerFactory(self._makePortal())
        args = dict(port=port, interface=interface, factory=pbd)
        self.addService('PB Server', internet.TCPServer, args)

class ServeupPowerup(powerup.Powerup):
    storeDependencies = [Serveup]

    powerupName = "Port Configuration"

    def addService(self, name, type, kwargs):
        return IServer(self.store.store).addService(name, type, kwargs)

    net = property(lambda self: IServer(self.store.store).net)



def addSomeKindOfServer(self,
                        interface = formless.String(),
                        port = formless.Integer()):
    pass

class IServerConfigurator(formless.TypedInterface):
    addHTTPServer = formless.autocallable(addSomeKindOfServer)
    addSMTPServer = formless.autocallable(addSomeKindOfServer)
    addIMAPServer = formless.autocallable(addSomeKindOfServer)
    addPOPServer = formless.autocallable(addSomeKindOfServer)
    addPBServer = formless.autocallable(addSomeKindOfServer)
    #servers = formless.List()


def addService(name):
    def addServiceNamed(self, interface, port):
        adder = getattr(
            IServer(self.original.store.store),
            'add%s' % name)
        adder(interface, port)
    return addServiceNamed


class ServeupPowerupDisplay(Base):
    __implements__ = IServerConfigurator, Base.__implements__

    def render_viewSwitcher(self, ctx, data):
        req = ctx.locate(inevow.IRequest)
        view = req.args.get('view', [None])[0]
        return tags.ul(_class="views")[
            tags.li(_class=view is None and "view-highlighted" or "view-normal")[
                tags.a(href=here())["existing servers"]
            ],
            [
                tags.li(_class=x == view and "view-highlighted" or "view-normal")[
                    tags.a(href=here(view=x))[x]
                ]
                for x in formless.IConfigurable(self).getBindingNames(ctx)
            ]
        ]

    def render_currentView(self, ctx, data):
        req = ctx.locate(inevow.IRequest)
        if req.args.has_key('view'):
            return webform.renderForms('', req.args['view'])
        return webform.renderForms('existingServers')

    document = taglib.page("Port Configuration")[
        render_viewSwitcher,
        render_currentView,
    ]

    addHTTPServer = addService('HTTPServer')
    addSMTPServer = addService('SMTPServer')
    addIMAPServer = addService('IMAPServer')
    addPOPServer = addService('POPServer')
    addPBServer = addService('PBServer')

    def configurable_existingServers(self, ctx):
        from quotient.web import pools
        rv = pools.PoolBrowser(list(self.original.net.queryIndex()))
        ctx.remember(rv, inevow.IRendererFactory)
        return rv

#    def servers():
#        def get(self):
#            return list(self.original.net.queryIndex())
#        return get,
#    servers = property(*servers())
