
from twisted.python.components import Interface
from twisted.python import log
from atop.powerup import Storeup, Powerup
from atop.store import Pool, Item
import formless

from atop.faraway import IFarIndex
from quotient.sysup import IControlPanel, ControlPanel, makeUser
from quotient.prefup import IPrefs

EXAMPLE_TICKET = 'glyph-3-awmzhtxtjbykroowfdrt'

from random import random, randint
from md5 import md5

def genTicket():
    "Generate some pretty garbage."
    return ''.join([chr(randint(97,122)) for x in range(20)])

class IBoxOffice(formless.TypedInterface):
    """A place where you go to buy your tickets and stuff.
    """

    def redeemTicket(self,
                     request = formless.Request(),
                     ticket = formless.String(
        label="Ticket",
        description="Enter your ticket information here.  For example, "+EXAMPLE_TICKET+"."
        "  (We suggest copying and pasting from your invitation, to avoid mistakes.)"),
                     desiredUsername=formless.String(
        label="Desired Username",
        description="Enter the username that you want to have when your account is created."),
                     desiredPassword=formless.Password(
        label="Desired Password",
        description="Enter the password you want to log in with.")
                     ):
        """Redeem a ticket from a string.
        """
    redeemTicket = formless.autocallable(redeemTicket)

class BoxOffice(Storeup):
    __implements__ = IBoxOffice, Storeup.__implements__

    postLocation = 'claim'

    def setUpPools(self, topstore):
        topstore.setComponent(IBoxOffice, self)

    def redeemTicket(self, request, ticket, desiredUsername, desiredPassword):
        vals = {'ticket': ticket,
                'desiredUsername': desiredUsername,
                'desiredPassword': desiredPassword}
        if not ticket or not desiredUsername or not desiredPassword:
            # XXX FIXME AUGH accomodate Nones from freeform: this is STUPID and
            # WRONG and EXACTLY THE KIND OF CODE THAT FREEFORM IS SUPPOSED TO
            # PREVENT ME FROM WRITING
            errs = {}
            for k, v in vals.items():
                if not v:
                    errs[k] = "Please enter a value."
            raise formless.ValidateError(errs, "You forgot some values.", vals)
            
        newusername = desiredUsername
        newpasswd = desiredPassword
        "runs on the login/registration server"
        from quotient.public import _domainFromHeader
        domain = _domainFromHeader(self.store, request)
        L = ticket.split('-')
        ticketProblem = formless.ValidateError({
                'ticket': 'Improperly formed ticket.'
                }, 'Please re-enter your ticket.', vals)
        try:
            if len(L)!=3:
                raise ValueError()
            issuer, ordinal, junk = L
            ordinal = int(ordinal)
        except ValueError:
            raise ticketProblem

        # when we're really going across multiple machines, this will be a
        # full-blown distributed transaction.  So how can we cheat to avoid
        # that level of complexity?  All we have to do *for the moment* is to
        # keep the user-creation and the ticket-checking within the same
        # transaction so that we either give the user their account or we
        # don't.  When we do go with full-blown distributed transactions, those
        # are the only two things that have to remain synchronized about this
        # whole process: once the new user's account has been created, the
        # ticket must remain consumed, and once the ticket has been consumed,
        # the new user's account must be created (or the ticket must be
        # refundable WITHOUT human intervention)

        d = IFarIndex(self.store, persist=False).farByName(issuer, domain).far(
            ITicketIssuer, "claimTicket", ordinal, junk, newusername, newpasswd, domain)
        # so, to make that above comment brief: if d fires, we are _in_
        def exceptBlock(f):
            f.trap(KeyError)
            raise ticketProblem

        def weAreIn(newUs):
            # NOT in a transaction!  Don't touch anything!
            from nevow.inevow import IResource
            s = request.getSession()
            res = IResource(newUs)
            s.setResourceForPortal(res, s.guard.resource.portal, res.logout)
            from quotient.web.root import afterPostView
            afterPostView(request, IPrefs(newUs))
        d.addCallbacks(weAreIn, exceptBlock)
        return d

class ITicketIssuer(formless.TypedInterface):
    """Interface for Avatar that allows issuing tickets.
    """

    def issueTicket(self):
        """Issue a new ticket for another person to sign up.
        """

    issueTicket = formless.autocallable(issueTicket)
    issuedTickets = formless.List()
    redeemedTickets = formless.List()
    remainingTicketCount = formless.Integer(immutable=True)


class InvalidTicket(formless.InputError):
    """The user passed an invalid ticket.  Whoops!
    """
    def __init__(self, *what):
        formless.InputError.__init__(self, "That ticket is invalid.  Try copying and pasting it?")

class NoMoreTickets(formless.InputError):
    """You don't have any more tickets available to give out.
    """
    def __init__(self, *what):
        formless.InputError.__init__(self, "You don't have any more tickets to give out.")

class TicketIssuer(Powerup):
    __implements__ = Powerup.__implements__, ITicketIssuer
    remainingTicketCount = 0
    powerupName = "Ticket Issuer"

    storePowerups = [ControlPanel]

    def setUpPools(self, avatar):
        avatar.setComponent(ITicketIssuer, self)
        self.redeemedTickets = {}
        self._issuedTickets = []
        try:
            itc = self.store.store.service.parent.initialTicketCount
            self.remainingTicketCount = itc
        except:
            pass #log.err()

    def claimTicket(self, ordinal, junk, newusername, newpassword, domain):
        """Claim a ticket based on an ordinal and some hash junk.

        Produce a new user.

        @type ordinal: C{int}
        @type junk: C{str}
        @type newusername: C{str}
        @type newpassword: C{str}
        @param newusername: the preferred username of the new user.
        @param newpassword: the preferred password of the new user.

        @return: new Avatar object with password set
        """
        try:
            storedJunk = self._issuedTickets[ordinal]
        except IndexError:
            raise InvalidTicket("ordinal too large")
        else:
            if storedJunk == junk:
                if self.redeemedTickets.has_key(ordinal):
                    raise InvalidTicket("That ticket was already redeemed.")
                ### okay, now _THIS_ part should move into a separate function at some point
                cp = IControlPanel(self.store.store)
                av = makeUser(self.store.store, newusername, newpassword, cp.POWERUPS, False)
                p = IPrefs(av)
                p.setPref('realname', newusername)
                p.setPref('replyto', newusername+'@'+domain)
                self.touch()
                self.redeemedTickets[ordinal] = newusername
                return av
            else:
                raise InvalidTicket("hash mismatch %r != %r" % (storedJunk, junk))

    def issueTicket(self):
        """Issue a ticket for copying to another person (and return a string to
        be displayed to the user).
        """
        # TODO: talk to the provisioner about this!
        if self.remainingTicketCount <= 0:
            raise NoMoreTickets()
        self.touch()
        self.remainingTicketCount -= 1
        tick = genTicket()
        ident = len(self._issuedTickets)
        self._issuedTickets.append(tick)
        return 'Issued ticket %s-%s-%s.' % (self.store.name, ident, tick)

    def issuedTickets():
        def get(self):
            return ['-'.join((self.store.name, str(num), val)) for num, val in enumerate(self._issuedTickets)
                    if not self.redeemedTickets.has_key(num)]
        def set(self, value):
            raise AttributeError("You can't set that.")
        return get, set

    issuedTickets = property(*issuedTickets())
