# -*- test-case-name: quotient.test.test_people -*-

import rfc822
import time

from twisted.python.components import Interface
from twisted.python import log

from atop.powerup import Powerup
from atop.store import Item, Pool, ItemForwardingComponent, ItemReferenceIndex
from atop.store import IPipe, NameIndex, INDEX_AND_ITEM, CaseInsensitiveIndex
from atop.store import referenced, indexed, registerIndex
from atop.upgrade import versionedUpgrade

from quotient.mimemessage import IMIMEDelivery, ICleanPoint, EmailPowerup

from quotient import provisioner
from quotient.smarts.base import ExtractBase, ExtractCollectorBase
from quotient.addressbook import AddressBook
from quotient.smartenup import ISmartenup, Smartenup, retroactivelyExtract
from quotient.prefup import IPrefs
from quotient.application import IPeopleApp, PeopleApp


class TabularCard:
    """Persistence Hack
    """

class Card:
    """A contact-information holding class; a metaphorical calling card.
    """
    name = None
    def __init__(self, person, title, organization):
        raise RuntimeError("Don't instantiate cards!  This class is just here for persistence.")
        self.person = person
        self.title = title
        self.organization = organization
        self.emails = {}
        self.phones = {}
        self.name = "%s %s" % (title, organization)
        self.dateCreated = time.gmtime(time.time())


from twisted.persisted.styles import Versioned

LAST_NAME = registerIndex(str, 'lastName')
FIRST_NAME = registerIndex(str, 'firstName')
_default = object()
class Person(Item, Versioned):
    firstName = indexed(FIRST_NAME)
    lastName = indexed(LAST_NAME)
    
    persistenceVersion = 3

    def upgradeToVersion3(self):
        self.touch()
        self.firstName, self.lastName = self.lastName, self.firstName

    def upgradeToVersion2(self):
        self.touch()
        sname = self.name.split()
        self.lastName = sname[-1]
        self.firstName = sname[0]
        for label, addr in self.emails.iteritems():
            self.setEmailAddress(addr, label)

    def upgradeToVersion1(self):
        """Upgrade from card-oriented person to non-card-oriented one.
        """
        self.touch()
        self.emails = {}
        self.phones = {}
        for card in self.cards:
            self.emails.update(card.emails)
            self.phones.update(card.phones)
        del self.cards

    def _updateDictAndIndex(self, dicti, index, label, value):
        self.touch()
        previous = dicti.get(label)
        if value is None and previous is not None:
            del dicti[label]
            if not previous in dicti.values():
                self.removeFromIndex(index, previous)
        else:
            dicti[label] = value
        self.appendToIndex(index, value)

    def _removeFromDictAndIndex(self, dicti, index, label):
        self.touch()
        previous = dicti.pop(label)
        if not previous in dicti.values():
            self.removeFromIndex(index, previous)

    def setName(self, firstName, lastName):
        self.touch()
        self.firstName = firstName
        self.lastName = lastName
        self.name = firstName + ' ' + lastName

    def setEmailAddress(self, address, label=''):
        self._updateDictAndIndex(self.emails, 'email', label, address.lower())

    def setPhoneNumber(self, number, label=''):
        self._updateDictAndIndex(self.phones, 'phones', label, number)

    def recentMessages(self, limit=5):
        """Returns an iterator.
        """
        mp = IPersonCollector(self.store).getMessagesPool()
        if mp is not None:
            rt = self.referenceTo()
            return mp.queryIndex('person', startKey=rt,
                                 endKey=rt, backwards=True, limit=limit)
        return iter([])

    def __init__(self, store, name, creator=None):
        Item.__init__(self, store, name)
        if creator:
            self.creatorRef = creator.getRef()
        else:
            me = IPrefs(store).me
            if me is not None:
                self.creatorRef = me.referenceTo()
            else:
                self.creatorRef = None
        self.emails = {}
        self.phones = {}

    def histogram(self, days=1000):
        msgs = IPersonCollector(self.store).getMessagesPool()
        D = {}
        now = time.gmtime()
        then = time.mktime(now) - (60 * 60 * 24 * days)
        selfref = self.getRef()
        for m in msgs.queryIndex('received',  endKey=then, backwards=True):
            if m.creatorRef == selfref:
                y, m, d = m.dateReceived[:3]
                day = (y,m,d)
                dayMsgs = D.setdefault(day,0)
                D[day] = dayMsgs + 1
        days = D.keys()
        days.sort()
        return [(day, D[day]) for day in days]


class PersonExtract(ExtractBase):
    pass


class IPersonCollector(Interface):
    """The address book.
    """


class PersonCollector(ExtractCollectorBase):
    __implements__ = IPersonCollector, IPipe, Powerup.__implements__, \
                     provisioner.ProvisionedPowerup.__implements__, provisioner.IProvisionable
    powerupName = 'People'
    dependencies = [AddressBook, Smartenup, PeopleApp]

    persistenceVersion = 4

    oldPeoplePool = referenced

    def upgradeToVersion4(self):
        self.touch()
        app = IPeopleApp(self.store)
        app.appRoot.removeItem(self.poolOfExtracts)
        smarts = ISmartenup(self.store)
        smarts.smartsPool().removeItem(self.poolOfExtracts)
        self.oldPeoplePool = self.poolOfExtracts
        self.setupExtractPool(self.poolOfExtracts.name)
        self.setupPeoplePool(self.poolOfExtracts)
        app.appRoot.addItem(self.poolOfExtracts)
        smarts.smartsPool().addItem(self.poolOfExtracts)
        pup, d = retroactivelyExtract(self.store, PersonCollector)
        d.addErrback(log.err)

    def upgradeToVersion3(self):
        self.touch()
        versionedUpgrade(self.getPeoplePool())

    def upgradeToVersion2(self):
        self.touch()
        pp = self.getPeoplePool()
        pp.removeIndex('name')
        versionedUpgrade(pp)

    def upgradeToVersion1(self):
        self.touch()
        from quotient.prefup import IPrefs
        ignore = IPrefs(self.store)
        app = IPeopleApp(self.store)
        self.extractPool().name = 'Everybody'
        app.appRoot.addItem(self.extractPool())
        self.store.getRootPool().removeItem(self.extractPool())

    def setUpPools(self, store):
        ExtractCollectorBase.setUpPools(self, store, name='Everybody', type=None)
        store.setComponent(IPersonCollector, self)
        self.getMessagesPool().addIndex(ItemReferenceIndex('person', store))
        self.setupPeoplePool(self.getPeoplePool())
        # Hack to get it accssible from root for left nav
        app = IPeopleApp(store)
        app.appRoot.addItem(self.extractPool())
        # for searching when we scan email
        self.invisible = False

    def setupPeoplePool(self, p):
        p.addIndex(NameIndex('email'))
        p.addIndex(CaseInsensitiveIndex(LAST_NAME.index))
        p.addIndex(CaseInsensitiveIndex(FIRST_NAME.index))

    def newPerson(self, newPersonName, guessed=False):
        newPersonName = newPersonName.strip()
        if not newPersonName:
            newPersonName = 'No_Name'
        p = Person(self.store, newPersonName)
        self.getPeoplePool().addItem(p)
        fName, lName = self.inferNames(newPersonName)
        p.firstName, p.lastName = self.inferNames(newPersonName)
        return p

    def inferNames(self, fullName):
        """
        Try to find infer a first name and last name from
        a string.
        It currently only handles names with a space in them
        e.g. John Keynes or J. Keynes
        and names with commas e.g.
        Keynes, John and Keynes, J.

        Names like J Keynes III, J Keynes Sr. will confuse it.
        It will also get names for things like "Family Newsletter List"
        If we can't infer a name we return '' as the first name
        and the original arguement as the last name.
        """
        if fullName.count(','):
            lName, fName = fullName.split(',', 1)
            return fName, lName
        elif fullName.count(' '):
            lName = fullName.split()[-1]
            fName = ' '.join(fullName.split()[:-1])
            return fName, lName
        else:
            return '', fullName
        

    def getPeoplePool(self):
        return self.extractPool()

    def getMessagesPool(self):
        return ICleanPoint(self.store).getItem().sourcePoolRef.getItem()

    def peopleFromMessage(self, item, hdr = 'from'):
        fromAddress = item[hdr]
        if not fromAddress:
            # No from header: let's not bother.
            return
        name, address = item.nameAndAddress()
        if not address:
            # Invalid header: let's not bother.
            return
        address = address.lower()
        folks = list(self.getPeoplePool().queryIndex('email', startKey=address, endKey=address))
        return folks

    def extractItem(self, item):
        if hasattr(item, 'creatorRef') and item.creatorRef:
            # Item already has a creator assigned
            return
        folks = self.peopleFromMessage(item)
        if not folks:
            # doesn't exist.  create!
            name, address = item.nameAndAddress()
            if not name:
                name = address
            person = self.newPerson(name)
            person.setEmailAddress(address)
        else:
            person = folks[-1]
        item.setIndexKey('person', person.referenceTo())
        item.creatorRef = person.getRef()

    def itemRemoved(self, item):
        people = self.peopleFromMessage(item)
        for person in people or ():
            try:
                self.getPeoplePool().removeItem(person)
            except KeyError:
                # No such person. Probably a multiple spammer
                pass

    def itemChanged(self, item, notification=None):
        pass
