# -*- test-case-name: quotient.test.test_extract -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import re

from atop import store

from quotient.smarts import base
from quotient.proto import twisted_sip

from quotient import taglib

from twisted.python import components

class PhoneNumberExtract(base.Textract):
    def getLink(self):
        ### TODO: Expedient temporary hack.
        dialer = twisted_sip.IContact(self.store)
        return "%s/freeform_post:placeCall?dialstring=%s" % (dialer.storeID, self.description[self.descBegin:self.descEnd])


sep = r'[ -\.]'
lead = '(?:\D|^)'
area = '(?:(\d{3})%s)?' % (sep,)
exch = '(?:(\d{3}))'
suff = '(?:(\d{4}))'
trail = '(?:\D|$)'
all = area + exch + sep + suff + trail
NorthAmericanPhoneNumber = re.compile(all)
del sep, lead, area, exch, suff, trail, all

class IPhonePowerup(components.Interface):
    pass

class PhoneNumberCollector(base.ExtractCollectorBase):
    __implements__ = (IPhonePowerup, base.ExtractCollectorBase.__implements__,
                      store.IPipe)

    powerupName = "Phone Numbers"

    ExtractType = PhoneNumberExtract

    def setUpPools(self, avatar):
        base.ExtractCollectorBase.setUpPools(self, avatar)
        self.phoneExpressions = [NorthAmericanPhoneNumber]
        avatar.setComponent(IPhonePowerup, self)

    def extractPart(self, item, text, part=None):
        i = 0
        for expr in self.phoneExpressions:
            for m in expr.finditer(text):
                i += 1
                begin, end = m.start(), m.end()
                e = self.createExtract(m.group(), item, text, begin, end, part)
                self.foundExtract(item, e)
                if i > self.maxExtractsPer:
                    return
