# -*- test-case-name: quotient.test.test_tdb -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


from twisted.python import components, reflect


from nevow import tags
from formless import webform
from nevow import inevow
from nevow import url
from nevow import guard

from twisted.python.components import Interface, CannotAdapt

from quotient.web import webinterfaces
from quotient.functional import unittest as funit
from quotient.web import filer
from quotient import application

from atop import credup

from urllib import urlencode


class ITimer(Interface):
    """Marker interface to represent the time this page started rendering.
    """

class ILinkGenerator(Interface):
    """A generator for hyperlinks.
    """

    def urlTo(self, item, request):
        """Return a URL to an item for a given request.
        """

import time
def startTimer(context, data):
    context.parent.remember(time.time(), ITimer)
    return ''


def showTimer(context, data):
    now = time.time()
    then = context.locate(ITimer)
    return tags.span(id='timer')["%0.2f ms" % ((now - then) * 1000)]


def linkToAvatarAspect(aspect, label):
    def getAspect(context, data):
        ava = context.locate(inevow.IResource).avatar
        return href(aspect(ava))[
            label
        ]
    return getAspect


def linkToPool(poolName, label=None):
    def getPool(context, data):
        ava = context.locate(inevow.IResource).avatar
        rootPool = ava.rootPoolRef.getItem()
        try:
            pool = rootPool.getIndexItem('name', poolName, None)
            if pool is None:
                return "The pool %s was not found" % poolName
            return href(pool) [
                label or poolName
            ]
        except AttributeError:
            return "There was a problem locating the pool %s." % poolName

    return getPool

def linkToPowerup(powerupName, label=None):
    def getPowerup(context, data):
        ava = context.locate(inevow.IResource).avatar
        rootPool = ava.rootPoolRef.getItem()
        powerups = rootPool.getIndexItem('name', 'powerups')
        it = powerups.getIndexItem('name', powerupName, None)
        if it is None:
            return "The powerup %s was not found" % powerupName
        return href(it)[
            comment['linkToPowerup'],
            label or powerupName
        ]
    return getPowerup


def getAttribute(name):
    return lambda context, data: getattr(data, name, "No attribute named %r found on %r." % (name, data))


def eitherOr(eitherer, orer):
    return lambda context, data: (
        getattr(data, eitherer, None) or
        getattr(data, orer, None) or
        "No attribute named %r or %r found on %r." % (eitherer, orer, data))


name = getAttribute('name')
anId = getAttribute('storeID')
nameOrId = eitherOr('name', 'storeID')




def containerList(context, data):
    labels = context.locate(application.IDivmodApp).labels
    if not getattr(data, 'pools', None) or isinstance(data, credup.Avatar):
        return ''
    def isDeletable(pool):
        if pool in labels:
            return tags.a(href=url.here.child(
                'freeform_post!!removeLabel').add('label', pool.storeID))[tags.img(
                    src='/images/icons/close.png')]
        return ''
    return tags.div(_class="container-list")[
        tags.div(_class="container-label")["Labels: "], tags.ul(_class="container-list-body")[[
            tags.li(_class="backreference", data=it)[
                isDeletable(it), href(it)[nameOrId]]
            for it in data.iterpools()]]]


def commandLineAction(context, data):
    try:
        from quotient import commander
        from atop import credup
        ava = context.locate(credup.IAvatar)
        commander.ICommand(ava)
    except components.CannotAdapt:
        return ''
    req = context.locate(inevow.IRequest)
    for slot in context._locateSlots('prompt'):
        slot["Command:"]
    pth = req.path.split('/')
    if req.path[-1] != '/' and pth[-1] != 'quotient_command_line':
        act = action="%s/quotient_command_line" % pth[-1]
    else:
        act = action="quotient_command_line"
    for form in context._locateKeys('form'):
        form(action=act)
    return context.tag


def pupImg(x, cur):
    if x.storeID == cur:
        D = 'tabs-logged-in-on'
    else:
        D = 'tabs-logged-in'
    return tags.img(src="/images/%s/%s.png" % (D, x.name))


def visiblePowerups(context, data):
    from quotient import prefup
    from quotient.application import IApplication
    req = context.locate(inevow.IRequest)
    ava = req.getSession().getLoggedInRoot().store # context.locate(credup.IAvatar)
    cur = data.storeID
    if components.implements(data, IApplication):
        cur = data.storeID
    elif hasattr(data, 'applicationRef'):
        cur = data.applicationRef.storeID
    try:
        return [
            href(x)(id=funit.getIdAttr(context, x.name))[
                tags.li[
                    pupImg(x, cur)
                ]
            ]
            for x in list(prefup.IPrefs(ava).getPrimaryNavPool().queryIndex())
        ]
    except components.CannotAdapt:
        return ''


def navigation(context, data):
    from quotient import prefup
    req = context.locate(inevow.IRequest)
    ava = getattr(req.getSession().getLoggedInRoot(),'store',None) # context.locate(credup.IAvatar)
    if ava is None:
        return ""

    dash = prefup.IPrefs(ava).getVisiblePool()

    try:
        dashid = int(req.prepath[-1])
    except ValueError:
        dashid = ''
    return (
        tags.ul(id="navigation", key="leftnav")[
            visiblePowerups,
            #href(dash)(id=dashid == dash.storeID and 'tab-highlighted' or 'leftnav.Control Panel', _class="primary-tab")[tags.li["Settings"]],
        ], 
        tags.a(id='logout', href="/"+guard.LOGOUT_AVATAR)["Logout"],
        tags.a(id="help-link", onclick="document.getElementById('help-box').style.display='block'")["Help"])

def recentResult(context, data):
    try:
        result = context.locate(inevow.IStatusMessage)
    except KeyError:
        result = None
    if result:
        result = str(result)
        try:
            hand = context.locate(inevow.IHand)
        except KeyError:
            hand = None
        if hand is not None:
            handstr = str(hand)
            # Try not to render <Foo instance at 0xDECAFBAD>
            if handstr[0] == '<' and handstr[-1] == '>':
                result += " The result was an instance of %s." % hand.__class__.__name__
            else:
                result += " The result was %r." % hand
        return tags.span(_id="recent-result")[
            result
        ]

    return tags.span(_id="recent-result")[
        tags.xml("&nbsp;")
    ]

def pageclass(context, data):
    page = context.locate(inevow.IResource)
    data = context.locate(inevow.IData)
    return tags.comment[ "\nPage class is: ", reflect.qual(page.__class__),
                         "\nData class is: ", reflect.qual(data.__class__),
                         "\nOriginal class is: ", reflect.qual(getattr(page,'original',None).__class__),
                         '\n']


def errordebugging(context, data):
    from pprint import pformat
    errs = context.locate(inevow.IFormErrors)
    de = context.locate(inevow.IFormDefaults)
    return tags.comment["\nFreeform Error Debugging:\n%s\nDefault debugging:\n%s\n" % (pformat(errs.errors), de.__dict__)]


class page:
    def __init__(self, title, anonymous=False, failure=False, skin="default", **kw):
        self.b = body = tags.invisible()

        skinner = reflect.namedAny('quotient.web.skins.%s' % skin)
        self.html = skinner.mainSkin(reflect.namedAny(__name__), title, body, anonymous, failure)

    def __getitem__(self,item):
        self.b[item]
        return self.html


def messageHeader(x):
    def getIt(data):
        return headerToUTF8(data[x] or ('<No %s>' % x))
    return getIt


def headers(*names):
    def _(context, data):
        t = tags.table(_class="header-block")
        for o in names:
            if callable(o):
                name = o.name
                dn = True
                F = o
            else:
                name = o
                dn = data[name]
                F = messageHeader(name)
            if dn:
                t[
                    tags.tr(_class="header-row")[
                        tags.td(_class="header-name")[name, ": "],
                        tags.td(_class="header-value")[F]]
                ]
        return t
    return _


def br(context, data):
    return tags.div(style="clear: both")[' ']


import sre
import sys

def headerToUTF8(s):
    # Man what the heck
    found = sre.findall(r'=\?[^?]+\?[QqBb]\?[^?]+\?=', s)
    if not found:
        return unicode(s, sys.getdefaultencoding(), "replace").encode("utf8")
    for item in found:
        _, charset, encoding, text, _ = item.split('?')
        if encoding in 'Qq':
            text = text.replace("_", "=20").decode('quopri')
        else:
            text = text.decode('base64')
        try:
            s = s.replace(item, text.decode(charset))
        except LookupError:
            return '<Unknown Encoding: %s>' % charset
    return s.encode('utf8')

class DefaultLinkGenerator:
    __implements__ = ILinkGenerator

    def urlTo(self, item, request):
        return '/me/%s' % item.storeID


class href(object):
    def __init__(self, item, *args, **kw):
        self.item = item
        self.args = args
        self.kw = kw
        self.callkw = {}
        self.getitems = []

    #def getComponent(self, interface, registry=None, default=None):
    #    if components.implements(self, interface):
    #        return self
    #    else:
    #        return MethodSerializer.getComponent(self, interface, registry, default)

    def getargs(self, request):
        return {}

    def original(self, context, data):
        request = context.locate(inevow.IRequest)
        linkgen = context.locate(ILinkGenerator)
        url = linkgen.urlTo(self.item, request)
        if url is None:
            return self.maketag(tags.span, _class='link-disabled')
        argz = self.getargs(request)
        argz.update(self.kw)
        path = '/'.join(map(str, (url,)+self.args))

        if argz:
            return self.maketag(tags.a, href='%s?%s' % (path, urlencode(argz, True)))
        else:
            return self.maketag(tags.a, href=path)

    def __call__(self, **kw):
        self.callkw.update(kw)
        return self

    def __getitem__(self, items):
        self.getitems.append(items)
        return self

    def maketag(self, proto, **kw):
        self.__call__(**kw)
        t = proto(**self.callkw)
        for gitem in self.getitems:
            t = t[gitem]
        return t


def HrefSerializer(original, context):
    from nevow.flat import serialize
    return serialize(original.original, context)


class stickyhref(href):
    """an href which carries the current request arguments along with it
    """
    def getargs(self, request):
        return request.args.copy()


## from nevow.renderer import render
## from nevow.context import WovenContext
## from nevow.inevow import ISerializable


def configurePage(*args, **kw):
    from formless import webform
    return page(*args, **kw)[ webform.renderForms() ]


def css(root):
    # just a stub to make sure this works; this just burns a bit of cpu right now,
    # it's not really parameterized
    skinner = reflect.namedAny('quotient.web.skins.default')
    return skinner.css(reflect.namedAny(__name__))


def images(root):
    skinner = reflect.namedAny('quotient.web.skins.default')
    return skinner.images(reflect.namedAny(__name__))


def missingImage():
    skinner = reflect.namedAny('quotient.web.skins.default')
    return skinner.missingImage(reflect.namedAny(__name__))

__all__ = ['page', 'configurePage', 'br', 'headers', 'messageHeader', 'headerToUTF8', 'href', 'missingImage']

