# -*- test-case-name: quotient.test.test_task -*-
# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os, time
from twisted.python import log
from twisted.python.components import Interface
from atop.store import IPipe, Pool, Item, NameIndex, DateIndex, IntIndex
from quotient.web.root import Base, TabularBase, column
from atop.powerup import Powerup, IPowerup

from quotient.searchup import Searchup, ISearch, IIndexable

from nevow.tags import *

from nevow import formless
from nevow import freeform

from quotient import taglib 
from quotient.prefup import IPrefs, Prefup

# self.setIndexKey('priority', 3)

class ITask(Interface):
    pass

class IGoal(Interface):
    pass

class IEditableGoal(formless.TypedInterface):
    def editGoal(
        self,
        name=formless.String(label='Name'),
    ):
        """
        """
    editGoal = formless.autocallable(editGoal)
    def webNewTask(
        self,
        req = formless.Request(),
        taskName=formless.String(label='Name'),
    ):
        """
        """
    webNewTask = formless.autocallable(webNewTask)


class Goal(Item):
    __implements__ = IGoal, IEditableGoal, IIndexable

    def __init__(self, store, summary):
        Item.__init__(self, store, summary)
        p = taskishPool(self.store, '%s Tasks' % summary)
        self.tasksPoolRef = p.getRef()
        self.number = 0

    def getTasksPool(self):
        return self.tasksPoolRef.getItem()

    def editGoal(self, name):
        self.touch()
        oldSumm = self.name
        self.name = name
        tm = ITaskmaster(self.store)

    def webNewTask(self, req, taskName):
        t = self.newTask(taskName)
        req.setComponent(freeform.IRedirectAfterPost, '../%s' % t.storeID)
        return t

    def newTask(self, taskName, owner=None, priority=3, origEst=0, est=0, elap=0, remain=0):
        # Most of this should probably move into TaskMaster
        tm = ITaskmaster(self.store)
        if owner is None:
            owner = ITaskmaster(self.store).nobodyRef.getItem()
        t = Task(self.store, taskName, self, owner, priority, origEst, est, elap, remain)
        owner.ownTask(t)
        self.getTasksPool().addItem(t)
        t.number = (self.number, t.index_uid(self.getTasksPool()))
        tm = ITaskmaster(self.store)
        tm.getTasksPool().addItem(t)
        return t

    def removeTask(self, task):
        tp = self.getTasksPool()
        tp.removeItem(task)
        task.ownerRef.getItem().removeTask(task)

    def indexableData(self):
        return {'text':self.name}

    def newGoal(self, name):
        tm = ITaskmaster(self.store)
        g = tm.newGoal(name)
        return g

    def __str__(self):
        return self.name


class GoalRow(TabularBase):
    def render_number(self, contact, data):
        return data.number
    column(render_number, True, 80)

    def render_name(self, contact, data):
        return taglib.href(data)[str(data.name)]
    column(render_name, True, 300)

    def render_openTasks(self, contact, data):
        tp = data.getTasksPool()
        return len(list(tp.queryIndex()))
    column(render_openTasks, True, 80)


class GoalDetail(Base):
    __implements__ = IGoal, Base.__implements__
    document = taglib.page(title="Goal Home")[
        p["Goal name: ", directive('name')],
        p['', directive('taskTable')],
        freeform.renderForms(),
    ]

    def render_name(self, context, data):
        return data.name

    def render_taskTable(self, context, data):
        from quotient.web import pools
        return pools.PoolBrowser(data.getTasksPool())

class OwnerType(formless.Typed):
    def coerceWithBinding(self, val, task):
        """Coerce a value with the help of an object, which is the object
        we are configuring.
        """
        val = val.strip()
        tm = ITaskmaster(task.store)
        peeps = tm.getOwnersPool()
        print peeps.getIndexKey('name', val)
        owner = peeps.getIndexItem('name', val, None)
        if owner is None:
            # No such owner
            return task.ownerRef.getItem()
        else:
            return owner


class IEditableTask(formless.TypedInterface):
    def editTask(
        self,
##        currentEstimate = formless.Choice(
##            ['1','2','3','4','5','6','7','8'],
##            label="Current estimate:",
##            description="Hours needed to complete tasks",),
        estimate = formless.Integer(label='Estimate'),
        elapsed = formless.Integer(label='Elapsed'),
        remain = formless.Integer(label='Remaining'),
        name = formless.Text(label='Name'),
        owner = OwnerType(label='Owner')
        ):
        """
        """
    editTask = formless.autocallable(editTask)

    def removeTask(self):
        """Delete this task from the system. This action can not be undone."""
    removeTask = formless.autocallable(removeTask)

class Task(Item):
    __implements__ = ITask, IEditableTask, IIndexable

    def __init__(self, store, name, goal, owner, priority,
                 originalEstimate=0, estimate=0, elapsed=0, remain=0, status='New'):
        Item.__init__(self, store, name)
        #self.journalPoolRef = Pool(store,'Task Journal')
        self.setOwner(owner)
        self.goalRef = goal.getRef()
        self.status = status
        self.priority = priority
        self.journal = [] # Changes, comments
        self.rank = 0     # For arbitrary sort order
        self.pre = []
        self.post = []
        self.originalEstimate = originalEstimate
        if estimate == 0:
            self.estimate = originalEstimate
        else:
            self.estimate = estimate
        self.elapsed = elapsed
        self.remain = remain

    def editTask(self, estimate, name, elapsed, remain, owner):
        self.name = name
        self.estimate = estimate
        self.elapsed = elapsed
        self.remain = estimate - elapsed
        self.touch()

    def setOriginalEstimate(self, est):
        self.originalEstimate = est
        self.estimate = est
        self.remain = est
        self.touch()

    def removeTask(self):
        tm = ITaskmaster(self.store)
        tm.removeTask(self)

    def split(self):
        tm = ITaskmaster(self.store)
        return tm.newTask("Split from " + self.name, owner=self.getOwner(), goal=self.goalRef.getItem())

    def index_priority(self, pool):
        return self.priority

    def index_lapsed(self, pool):
        return self.elapsed

    def index_remain(self, pool):
        return self.remain

    def index_owner(self, pool):
        return self.getOwner().name

    def index_estimate(self, pool):
        return self.estimate

    def getOwner(self):
        return self.ownerRef.getItem()

    def setOwner(self, owner):
        self.ownerRef = owner.getRef()

    def indexableData(self):
        return {'text':self.name}

    def __str__(self):
        return self.name

class TaskDetail(Base):
    __implements__ = ITask, Base.__implements__
    document = taglib.page(title="Task Home")[
        p["Task name: ", directive('name')],
        p["Owner: ", directive('owner')],
        p["Actions: ", directive('split')],
        freeform.renderForms(),
        ]

    def renderHTTP(self, request):
        if request.args.has_key('originalEstimate'):
            est = request.args['originalEstimate'][0]
            def _():
                self.original.setOriginalEstimate(int(est))
            self.original.store.transact(_)
            request.redirect(request.getHeader('referer'))
        elif request.args.has_key('split'):
            def _():
                t = self.original.split()
                request.redirect('/me/%s' % t.storeID)
            self.original.store.transact(_)
        else:
            self.doc = self.document
        return Base.renderHTTP(self, request)

    def render_name(self, context, data):
        return '%s.%s - %s' % (data.number[0], data.number[1], data.name)

    def render_owner(self, context, data):
        owner = data.getOwner()
        return p[taglib.href(owner)[owner.name]]

    def render_split(self, context, data):
        return taglib.href(data,split=data)['Split Task']


class TaskRow(TabularBase):
    """Represents the columns that display in each row
    of a table of tasks.
    """

    def render_number(self, context, data):
        return '%s.%s' % data.number
    column(render_number, True, 85)

    def render_goal(self, context, data):
        if data.goalRef is not None:
            goal = data.goalRef.getItem()
            return taglib.href(goal)[goal.name]
        else:
            return '--'
    column(render_goal, True, 85)

    def render_name(self, context, data):
        return taglib.href(data)[str(data.name)]
    column(render_name, True, 300)

    def render_priority(self, context, data):
        return str(data.priority)
    column(render_priority, True, 70)

    def render_owner(self, context, data):
        if data.ownerRef is not None:
            owner = data.getOwner()
            return taglib.href(owner)[owner.name]
        else:
            return '--'
    column(render_owner, True, 85)

    def render_estimate(self, context, data):
        est = data.estimate
        if est == 0:
            return [
                taglib.href(data, originalEstimate=((x == 8) and str(x) or (str(x)+'.'))) for x in [1, 2, 4, 6, 8]
            ]
        else:
            return est
    column(render_estimate, True, 70)

    def render_elapsed(self, context, data):
        return str(data.elapsed)
    column(render_elapsed, True, 70)

    def render_remain(self, context, data):
        return str(data.remain)
    column(render_remain, True, 70)

    def render_received(self, context, data):
        return str(data.remain)
    column(render_received, False, 70)

class IOwn(Interface):
    pass

class OwnerRow(TabularBase):
    def render_name(self, context, data):
        return taglib.href(self.original)[str(data.name)]
    column(render_name, True, 300)


class IEditableOwner(formless.TypedInterface):
    def newTask(
        self,
        name = formless.String(label='Task description'),
        ):
        """
        """
    newTask = formless.autocallable(newTask)


class OwnerDetail(Base):
    __implements__ = Base.__implements__, IOwn
    document = taglib.page(title="Owner Home")[
        p["Owner name: ", directive('name')],
        p["Tasks: ", directive('taskTable')],
        freeform.renderForms(),
    ]

    def render_name(self, context, data):
        return data.name

    def render_taskTable(self, context, data):
        from quotient.web import pools
        return pools.PoolBrowser(data.getTasksPool()) #, data.getTasksPool().queryIndex(), False)


class Owner(Item):
    __implements__ = IOwn, IEditableOwner

    def __init__(self, store, name):
        Item.__init__(self, store, name)
        p = taskishPool(store, name="%s's Tasks" % name)
        self.tasksPoolRef = p.referenceTo()

    def ownTask(self, task):
        self.tasksPoolRef.getItem().addItem(task)
        task.setOwner(self)

    def removeTask(self, task):
        tp = self.getTasksPool()
        if task in tp:
            tp.removeItem(task)
        else:
            raise 'Removing non-existent task from Owner?'

    def getTasksPool(self):
        return self.tasksPoolRef.getItem()

    def newTask(self, name):
        tm = ITaskmaster(self.store)
        tm.newTask(name, owner=self)

    def __repr__(self):
        return '<Owner:%s>' % self.name

    
class ITaskmaster(Interface):
    pass


class Taskmaster(Powerup):
    __implements__ = IPipe, ITaskmaster, Powerup.__implements__
    dependencies = [Searchup, Prefup]
    powerupName = 'Taskmaster'

    def setUpPools(self, store):
        store.setComponent(ITaskmaster, self)
        myRoot = Pool(store, name='Taskmaster')
        self.appRootRef = myRoot.getRef()
        store.getRootPool().addItem(myRoot)

        searcher = ISearch(store)

        o = Pool(store, name='All Owners')
        self.ownersPoolRef = o.getRef()
        myRoot.addItem(o)
        o.addIndex(NameIndex('name'))

        t = taskishPool(store, name='All Tasks')
        self.tasksPoolRef = t.getRef()
        myRoot.addItem(t)
        t.addIndex(NameIndex('name'))
        searcher.indexPool(t)

        d = Pool(store, name = 'Deleted Tasks')
        myRoot.addItem(d)
        self.deletedPoolRef = d.getRef()

        g = Pool(store, name='All Goals')
        self.goalsPoolRef = g.getRef()
        myRoot.addItem(g)
        g.addIndex(NameIndex('name'))
        searcher.indexPool(g)

        # create a Nobody to own the unowned tasks
        nb = Owner(store, 'Unassigned')
        o.addItem(nb)
        self.nobodyRef = nb.getRef()
        self.noGoalRef = self.newGoal('Unassigned').getRef()
        IPrefs(store).addToPrimaryNav(self)

    def newGoal(self, name, creator=None):
        g = Goal(self.store, name)
        self.getGoalsPool().addItem(g)
        g.number = g.index_uid(self.getGoalsPool())
        return g

    def newOwner(self, name):
        o = Owner(self.store, name)
        self.getOwnersPool().addItem(o)
        return o

    def newTask(self, name, owner=None, goal=None):
        if goal is None:
            goal = self.noGoalRef.getItem()
        if owner is None:
            owner = self.nobodyRef.getItem()
        task = goal.newTask(name, owner=owner)
        #self.getTasksPool().addItem(task)
        return task

    def removeTask(self, task):
        """Remove a task from the system.
        First remove it from the Goal.
        The Goal will also remote it from the Owner.
        """
        task.goalRef.getItem().removeTask(task)
        tp = self.getTasksPool()
        if task in tp:
            tp.removeItem(task)
            self.deletedPoolRef.getItem().addItem(task)
        else:
            raise 'Removing non-existent task from Taskmaster?', task

    def getOwnersPool(self):
        return self.ownersPoolRef.getItem()

    def getTasksPool(self):
        return self.tasksPoolRef.getItem()

    def getGoalsPool(self):
        return self.goalsPoolRef.getItem()

    def getAppRootPool(self):
        return self.appRootRef.getItem()

    def backup(self, filename='anviltasks.txt'):
        F = open(filename, 'wb')
        for g in self.getGoalsPool().queryIndex():
            for t in g.getTasksPool().queryIndex():
                F.write('%s\t%s\t%s\t%s\t' % (g.name,
                                                t.name,
                                                t.priority,
                                                t.ownerRef.getItem().name
                                              ))
                F.write('%s\t%s\t%s\t%s' % (t.originalEstimate,
                                              t.estimate,
                                              t.elapsed,
                                              t.remain
                                              ))
                F.write('\r\n')
        F.close()


    def restore(self, filename='anviltasks.txt'):
        tasksPool = self.getTasksPool()
        owners = {}
        def getOwner(oName):
            if not oName:
                return self.nobodyRef.getItem()
            oName = oName.strip()
            own = owners.get(oName, None)
            if not own:
                own = self.newOwner(oName)
                owners[oName] = own
            return own

        goals = {}
        def getGoal(gName):
            goal = goals.get(gName, None)
            if not goal:
                goal = self.newGoal(gName)
                goals[gName] = goal
            return goal

        if not os.path.exists(filename):
            return 'Error! No such file, %s' % filename

        for tl in open(filename,'rb').readlines():
            if tl:
                parts = tl.split('\t', 7)
                if len(parts) == 4:
                    ownerName, goalName, taskName, origEst = parts
                    est = elap = remain = priStr = 0
                elif len(parts) == 5:
                    ownerName, goalName, taskName, origEst, status = parts  # status is ignored for now
                    est = elap = remain = priStr = 0
                elif len(parts) == 8:
                    goalName, taskName, priStr, ownerName, origEst, est, elap, remain = parts
                else:
                    log.msg("ERROR in %s: %r" % (filename, tl))
                    continue
                try:
                    owner = getOwner(ownerName)
                    goal = getGoal(goalName)
                    t = goal.newTask(taskName, owner, int(priStr), int(origEst), int(est), int(elap),
                                     int(remain))
                except:
                    print parts
                    log.err()
        return 'Restore Successful.'


class IBasicTaskmaster(formless.TypedInterface):
    def newGoal(self,
            req=formless.Request(),
            name=formless.String(label='Name'),):
        """Create a new Goal"""
    newGoal = formless.autocallable(newGoal)

    def backup(self):
        """Back up data to anviltasks.txt"""
    backup = formless.autocallable(backup)

    def restore(self):
        """Restore data from anviltasks.txt"""
    restore = formless.autocallable(restore)


class TaskmasterDisplay(Base):
    __implements__ = IBasicTaskmaster, Base.__implements__

    def pools(self, context, data):
        from quotient.web import pools
        return pools.PoolBrowser(self.original.getAppRootPool())

    document = taglib.page(title="Task Manager Configuration")[
        pools,
        freeform.renderForms()
    ]

    def backup(self):
        self.original.backup()
        return "Backed up!"

    def restore(self):
        self.original.restore()
        return "Backed up!"

    def newGoal(self, req, name):
        gl = self.original.newGoal(name)
        req.setComponent(freeform.IRedirectAfterPost, "../%s" % gl.storeID)


def taskishPool(avatar, name):
    pool = Pool(avatar, name=name)
    pool.addIndex(NameIndex('name'))
    pool.addIndex(NameIndex('owner'))
    pool.addIndex(IntIndex('priority'))
    pool.addIndex(IntIndex('estimate'))
    pool.addIndex(IntIndex('elapsed'))
    pool.addIndex(IntIndex('remain'))
    pool.suggestView('Table')
    pool.suggestIndex('priority', backwards=False)
    return pool

##def sortedQuery(self, index, getsortkey):
##    L = []
##    lastKey = None
##    for k, q in self.queryIndex(index, result=INDEX_AND_ITEM):
##        if k != lastKey:
##            L.sort()
##            for sk, r in L:
##                 yield r
##             lastKey = k
##             L = []
##         j = getsortkey(q)
##         L.append((j, q))
