# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from os.path import join as opj

from atop.powerup import IPowerStation
from atop.sched import SystemScheduler, IScheduler
from atop.test import unittest

from quotient.test import emaildata, helper
from quotient.mimemessage import MIMEMessageReceiver, messageFromStructure as mfs
from quotient.cleanup import Ruler, Whitelist
from quotient.proto.smtpout import IOutgoingMail

class YesRule:
    def match(self, item):
        return True

class NoRule:
    def match(self, item):
        return False

class TestRuler(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [SystemScheduler]

    def dbSetUp(self):
        self.substore = self.createAvatar(())
        self.ruler = IPowerStation(self.substore).installPowerup(Ruler)
        self.parser = MIMEMessageReceiver(self.substore, self.noop)
        self.msgLite = mfs(self.substore,
                  [('To', 'clam@bake.org, slice@pizza.tv'),
                   ('cc', 'bingo@night.net'),
                   ('bcc', 'spiral@bound.org'),
                   ('Subject',  'tst'),
                   ('From',  'amir@bar.net')],
                  'this is the text')

    def dbTearDown(self):
        # Clear out pending calls
        dc = IScheduler(self.store).nextCall
        dc and dc.cancel()

    def dbtestYes(self):
        self.ruler.addRule('Yes', YesRule())
        res = self.ruler.match(self.msgLite)
        self.failUnless(res, 'Expected True, got %s' % res)

    def dbtestNo(self):
        self.ruler.addRule('Yes', NoRule())
        res = self.ruler.match(self.msgLite)
        self.failIf(res, 'Expected False, got %s' % res)
        
    def noop(self, item):
        pass

    def testWhitelist(self):
        wl = Whitelist()
        wl.addToList('More@test.strings')
        self.failUnless(len(wl.list) == 1, 'Expected 1, got %s' % len(wl.list))
        m = emaildata.test_messages[0]
        msg = self.parser.feedStringNow(m)
        self.failUnless(wl.match(msg) == True, 'Expected True, got %s' % wl.match(msg))

    def dbtestWhitelistAndRuler(self):
        wl = Whitelist()
        wl.addToList('More@test.strings')
        self.ruler.addRule('WL', wl)
        m = emaildata.test_messages[0]
        msg = self.parser.feedStringNow(m)
        self.failUnless(self.ruler.match(msg), 'Expected True, got %s' % self.ruler.match(msg))

    def dbtestWhitelistAndRulerFalse(self):
        wl = Whitelist()
        wl.addToList('zo0ooozoooooo@test.strings')
        self.ruler.addRule('WL', wl)
        m = emaildata.test_messages[0]
        msg = self.parser.feedStringNow(m)
        self.failIf(self.ruler.match(msg), 'Expected False, got %s' % self.ruler.match(msg))
        
    def dbtestWhitelistAndRulerFalseUp(self):
        wl = Whitelist()
        wl.addToList('zo0ooozoooooo@test.strings')
        wl.addToList('More@test.strings')
        self.ruler.addRule('WL', wl)
        m = emaildata.test_messages[0]
        msg = self.parser.feedStringNow(m)
        self.failUnless(self.ruler.match(msg), 'Expected True, got %s' % self.ruler.match(msg))

    def dbtestWhitelistAndRulerFalseNo(self):
        self.ruler.addRule('No', NoRule())
        wl = Whitelist()
        wl.addToList('zo0ooozoooooo@test.strings')
        wl.addToList('More@test.strings')
        self.ruler.addRule('WL', wl)
        m = emaildata.test_messages[0]
        msg = self.parser.feedStringNow(m)
        self.failUnless(self.ruler.match(msg), 'Expected True, got %s' % self.ruler.match(msg))

    def dbtestWhitelistAuto(self):
        self.outmail = IOutgoingMail(self.substore)
        self.outmail.sentRef.getItem().addItem(self.msgLite)
        m = emaildata.test_messages[0]
        msg = self.parser.feedStringNow(m)
        L = self.ruler.whitelist.list
        self.assertEquals(len(L), 4, 'Expected 4, got %s' % len(L))
