
import string

from atop.test import unittest
from atop.store import Pool, NameIndex
from atop import store

from quotient.proto.mboxlist import listMailboxes
from quotient.mimemessage import EmailPool

class GoodPool(Pool):
    def __init__(self, *a, **kw):
        Pool.__init__(self, *a, **kw)
        self.addIndex(NameIndex())
        self.addIndex(NameIndex('imapname'))

class ListMailboxesTestCase(unittest.TestCase):

    def setUpClass(self):
        store.MAX_LOCKS = 750

    def tearDownClass(self):
        store.MAX_LOCKS = 1000

    def txnwork(self, names):
        root = self.store.transact(GoodPool, self.store)
        for n in names:
            def _():
                inner = GoodPool(self.store, n)
                root.addItem(inner)
                return inner
            inner = self.store.transact(_)
            def _():
                for m in names:
                    doubleInner = EmailPool(self.store, m)
                    inner.addItem(doubleInner)
            self.store.transact(_)
        return root

    def vrfywork(self, root, names):
        mboxes = list(listMailboxes(self.store, root))
        expected = sum([
            ([first] +  [first + '/' + second for second in names])
            for first in names
            ], [])
        self.assertEquals(len(mboxes), len(expected))
        self.assertEquals([name for (name, pool) in mboxes], expected)

    def testJustWorks(self):
        root = self.txnwork('abc')
        self.vrfywork(root, 'abc')

    def testHugeFolderList(self):
        names = string.digits + string.uppercase + string.lowercase
        root = self.txnwork(names)
        self.vrfywork(root, names)
