# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from os.path import join as opj
import os
import cStringIO
from itertools import izip

from twisted.cred.portal import IRealm, Portal
from twisted.cred.checkers import ICredentialsChecker
from twisted.protocols.pop3 import IMailbox, POP3
from twisted.internet import protocol
from twisted.internet import defer
from twisted.persisted.styles import Versioned

from atop.store import Store, SubStore, Item, Pool, StorageService, IPipe, INDEX
from atop.powerup import IPowerStation
from atop.sched import IScheduler, SchedulerService, SubScheduler
from atop.test import unittest

from quotient import mimemessage
from quotient.mimemessage import MIMEMessageReceiver, IMIMEDelivery, MIMEMessage
from quotient import grabbers
from quotient.test import emaildata, helper

# from email.Message import Message as PyEmailMessage
from email import message_from_string

from quotient.test.emaildata import test_messages

class POP3TestMailbox:
    __implements__ = IMailbox
    def listMessages(self, index=None):
        if index is None:
            return map(len, test_messages)
        else:
            return len(test_messages[index])
    def getMessage(self, index):
        return cStringIO.StringIO(test_messages[index])
    def getUidl(self, index):
        return str(index)
    def deleteMessage(self, index):
        return
    def undeleteMessages(self):
        return
    def sync(self):
        return

class POP3TestRealm:
    __implements__ = IRealm
    def requestAvatar(self, avatarId, mind, *interfaces):
        return interfaces[0], POP3TestMailbox(), lambda : None

from twisted.internet import protocol

class POP3TestFactory(protocol.Factory):
    def __init__(self, portal):
        self.portal = portal

    def buildProtocol(self, addr):
        p3 = POP3()
        p3.portal = self.portal
        return p3

from twisted.cred import credentials

class POP3TestChecker:

    __implements__ = ICredentialsChecker
    credentialInterfaces = (credentials.IUsernamePassword,
                            credentials.IUsernameHashedPassword)

    def requestAvatarId(self, credentials):
        return 'anybody'

class MIMEMessageTestCase(unittest.TestCase, helper.ProvisionerMixin):

    def setUp(self):
        dbpath = self.mktemp()
        self.store = Store(opj(dbpath, "db"), opj(dbpath, "files"))
        StorageService(self.store)
        def _():
            self.svc = SchedulerService(self.store)
            self.substore = self.createAvatar([SubScheduler])
            self.svc.startService()
        self.store.transact(_)

    def tearDown(self):
        self.svc.stopService()
        self.store.close()

    def testBasicParsing(self):
        L = []
        for m in emaildata.test_messages:
            MIMEMessageReceiver(self.substore, L.append).feedStringNow(m)
        self.assertEquals(len(L),
                          len(emaildata.test_messages))
        L2 = [message_from_string(M) for M in emaildata.test_messages]

        for OurM, TheirM in zip(L, L2):
            self.failUnless(hasattr(OurM,'size'))
            if (len(OurM.headers) - 1) != len(TheirM._headers):
                # account for "x-divmod-processed" header.
                from pprint import pformat
                ourh = OurM.headers
                theirh = TheirM._headers
                ourh.sort()
                theirh.sort()
                self.fail("Header mismatch: \n %s\n\n %s" % (pformat(ourh), pformat(theirh)))
            if OurM.is_multipart():
                self.assertEquals(OurM.getPartByID(1), OurM.children[0])
                self.assertEquals(OurM.getPartByID(0), OurM)
                self.assertEquals(1, OurM.children[0].mimeID)
                self.assertEquals(0, OurM.mimeID)
            for ourPart, theirPart in izip(OurM.walk(), TheirM.walk()):
                self.assertEquals(
                    ourPart['subject'],
                    theirPart['subject'])
                self.assertEquals(
                    ourPart.is_multipart(),
                    theirPart.is_multipart())
                if not ourPart.is_multipart():
                    ourBody = ourPart.get_payload(decode=True)
                    theirBody = theirPart.get_payload(decode=True)
                    if ourBody != theirBody:
                        if ourBody[-1] == '\n':
                            if ourBody[:-1] == theirBody:
                                continue
                        raise '?'
                else:
                    self.assertEquals(len(ourPart.children),
                                      len(theirPart._payload))

    def testReceivingMessages(self):
        from quotient import grabbers
        from atop import powerup
        from twisted.internet import reactor
        prtl = Portal(POP3TestRealm())
        prtl.registerChecker(POP3TestChecker())
        p3f = POP3TestFactory(prtl)
        prt = reactor.listenTCP(0, p3f, interface="127.0.0.1")
        portnum = prt.getHost().port
        # reactor.connectTCP("127.0.0.1", portnum, )
        def _():
            pup = powerup.IPowerStation(self.substore).installPowerup(grabbers.GrabberPowerup)
            grab = pup.addPopGrabber("anybody", "anybody", "127.0.0.1", (portnum, False), "anybody@localhost", 100)
            return grab
        g = self.substore.transact(_)
        d = self.substore.transact(g.run)
        d.addCallback(lambda x: reactor.crash())
        reactor.run()
        def _2():
            # whiiiiiite boooooooox
            # powerup.IPowerStation(self.substore).installPowerup(mimemessage.EmailPowerup)
            arrivalPool = IMIMEDelivery(self.substore).arrivalRef.getItem()
            self.assertEquals(len(list(arrivalPool.queryIndex("uid"))),
                              len(test_messages))
            p2 = mimemessage.EmailPool(self.substore, "testing")
            for message in arrivalPool.queryIndex():
                p2.addItem(message)
            self.assertEquals(len(list(p2.queryIndex('sender'))),
                              len(test_messages))
        self.substore.transact(_2)

class GrabberSchedulingTestCase(unittest.TestCase, helper.ProvisionerMixin):

    def dbSetUp(self):
        self.svc = SchedulerService(self.store)
        self.substore = self.createAvatar([grabbers.GrabberPowerup])
        self.gp = grabbers.IGrabberAdder(self.substore)
        self.svc.startService()

    def dbTearDown(self):
        self.svc.stopService()

    def dbtestPausing(self):
        f = self.gp.addPopGrabber("user", "pass", "host", (12345, False), "user@host", 0)
        f.paused = True

        when = IScheduler(self.store).clock()
        d = self.store.transact(f.run)
        self.assertEquals(f.paused, True)
        self.assertEquals(d, None)
        self.assertEquals(f.currentProtocol, None)

    def dbtestMultipleRuns(self):
        from twisted.internet import reactor
        sf = ServerFactory()
        sf.protocol = protocol.Protocol
        p = reactor.listenTCP(0, sf).getHost().port

        cf = self.gp.addPopGrabber("user", "pass", "localhost", (p, False), "user@localhost", 0)
        d1 = cf.run()
        d2 = cf.run()
        self.assertIdentical(d1, d2)

        unittest.deferredResult(sf.connected)
        self.failUnless(sf.connMade)

class FlaggedItem(Item):
    def __init__(self, store, flags):
        Item.__init__(self, store)
        self.flags = flags
    def index_flags(self, pool):
        return self.flags

class Flags(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = self.createAvatar(())
        self.pool = Pool(self.avatar)
        self.pool.addIndex(mimemessage.FlagsIndex())

    def dbtestFlagCounter(self):
        a = self.pool.addItem
        a(FlaggedItem(self.avatar, 0))
        a(FlaggedItem(self.avatar, 1))
        a(FlaggedItem(self.avatar, 1 | 2))
        a(FlaggedItem(self.avatar, 2 | 4))
        a(FlaggedItem(self.avatar, 1 | 4))
        a(FlaggedItem(self.avatar, 1 | 8))
        a(FlaggedItem(self.avatar, 4 | 16))

        f = mimemessage.numberFlagged

        self.assertEquals(f(self.pool, 1), 4)
        self.assertEquals(f(self.pool, 2), 2)
        self.assertEquals(f(self.pool, 4), 3)
        self.assertEquals(f(self.pool, 8), 1)
        self.assertEquals(f(self.pool, 16), 1)
        self.assertEquals(f(self.pool, 0), 0)
        self.assertEquals(f(self.pool, 31), 6)

class EmailPool(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = self.createAvatar(())
        self.pool = mimemessage.EmailPool(self.avatar,'TestEmailPool')

    def dbtestNewCount(self):
        self.assertEquals(self.pool.newItemCount(), 0)
        self.pool.setUIDMark()
        self.assertEquals(self.pool.newItemCount(), 0)

class ServerFactory(protocol.ServerFactory):
    connMade = False
    def __init__(self):
        self.connected = defer.Deferred()

    def buildProtocol(self, addr):
        self.connMade = True
        self.connected.callback(self)
        return protocol.ServerFactory.buildProtocol(self, addr)

class SimulatedException(Exception):
    pass

class AtopInteraction(unittest.TestCase):
    def testReversionIncludesChildren(self):
        msg = mimemessage.MIMEMessage()
        msg.newChild()

        def addToStore():
            msg.addToStore(self.store)
        self.store.transact(addToStore)

        def mangle():
            msg.touch()
            raise SimulatedException("Simulated!")

        try:
            self.store.transact(mangle)
        except SimulatedException, e:
            pass
        else:
            self.failIf(True, "Expected exception not raised")

        self.assertIdentical(msg.children[0].parent, msg)


class TestPipe(Versioned):
    __implements__ = (IPipe,)

    persistenceForgets = ['adds', 'removes', 'changes']

    def __init__(self):
        self.adds, self.removes, self.changes = [], [], []
    def itemAdded(self, *a):
        self.adds.append(a)
    def itemRemoved(self, *a):
        self.removes.append(a)
    def itemChanged(self, *a):
        self.changes.append(a)

class TestPool(Pool):
    def __init__(self, *a, **kw):
        Pool.__init__(self, *a, **kw)
        self.addIndex(mimemessage.FlagsIndex())
        self.pipe = TestPipe()
        self.addPipe(self.pipe)

class Flags(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = self.createAvatar(())
        # Set a message and put it in two pools
        self.msg = mimemessage.messageFromStructure(
            self.avatar,
            [("Irrelevant Header", "Irrelevant Value")],
            "Body text")
        self.pools = [TestPool(self.avatar), TestPool(self.avatar)]
        map(Pool.addItem, self.pools, [self.msg] * 2)
        self.pools[0].pipe.adds = []
        self.pools[1].pipe.adds = []

    def dbtestPerMessageFlagChange(self):
        # Both pipes should see this
        self.msg.changeFlags(None, MIMEMessage.SEEN, True)
        for p in self.pools:
            p = p.pipe
            self.assertEquals(p.changes,
                              [(self.msg, {'flags': MIMEMessage.SEEN, 'oldFlags': 0})])
            p.changes = []

        # Make sure the indexes are correct as well
        q = self.pools[0].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [MIMEMessage.SEEN])
        q = self.pools[1].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [MIMEMessage.SEEN])

        # Both pipes should see this too
        self.msg.changeFlags(None, MIMEMessage.SEEN, False)
        for p in self.pools:
            p = p.pipe
            self.assertEquals(p.changes,
                              [(self.msg, {'flags': 0, 'oldFlags': MIMEMessage.SEEN})])
            p.changes = []

        # And check the indexes again
        q = self.pools[0].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [0])
        q = self.pools[1].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [0])

        for p in self.pools:
            self.assertEquals(p.pipe.adds, [])
            self.assertEquals(p.pipe.removes, [])

    def dbtestPerPoolFlagChanges(self):
        # This changes is only for a single pool, so only one pipe should see it
        self.msg.changeFlags(self.pools[0], MIMEMessage.DELETED, True)
        self.assertEquals(self.pools[0].pipe.changes,
                          [(self.msg, {'flags': MIMEMessage.DELETED, 'oldFlags': 0})])
        self.pools[0].pipe.changes = []
        self.assertEquals(self.pools[1].pipe.changes, [])

        # Make sure it is only changed in one index
        q = self.pools[0].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [MIMEMessage.DELETED])
        q = self.pools[1].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [0])

        # Now change it in the other pool
        self.msg.changeFlags(self.pools[1], MIMEMessage.DELETED, True)
        self.assertEquals(self.pools[0].pipe.changes, [])
        self.assertEquals(self.pools[1].pipe.changes,
                          [(self.msg, {'flags': MIMEMessage.DELETED, 'oldFlags': 0})])
        self.pools[1].pipe.changes = []

        # Make sure both indexes have the flag now
        q = self.pools[0].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [MIMEMessage.DELETED])
        q = self.pools[1].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [MIMEMessage.DELETED])

        # Unset it from the first pool
        self.msg.changeFlags(self.pools[0], MIMEMessage.DELETED, False)
        self.assertEquals(self.pools[0].pipe.changes,
                          [(self.msg, {'flags': 0, 'oldFlags': MIMEMessage.DELETED})])
        self.pools[0].pipe.changes = []
        self.assertEquals(self.pools[1].pipe.changes, [])

        # And make one last check of the indexes
        q = self.pools[0].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [0])
        q = self.pools[1].queryIndex('flags', result=INDEX)
        self.assertEquals(list(q), [MIMEMessage.DELETED])

        for p in self.pools:
            self.assertEquals(p.pipe.adds, [])
            self.assertEquals(p.pipe.removes, [])
