# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from os.path import join as opj

from atop.store import Store, SubStore
from atop.powerup import IPowerStation
from atop.test import unittest

from quotient.prefup import Prefup, IPrefs

from nevow.inevow import IResource

class TestPrefup(unittest.TestCase):

    def dbSetUp(self):
        self.substore = SubStore(self.store)
        self.prefs = IPowerStation(self.substore).installPowerup(Prefup)
        self.prefs.testPref = 'Divmod'

    def testBasic(self):
        self.failUnless(self.prefs.testPref == 'Divmod', 'I hope we did not fail this one!')

    def part1SetGet(self):
        del self.prefs

        prefs = IPrefs(self.substore)
        prefs.timezone = 'foo'
        prefs.touch()
    
    def part2SetGet(self):
        self.assertEquals(IPrefs(self.substore).timezone, 'foo')

    def part3SetGet(self):
        prefs = IPrefs(self.substore)
        IPrefs(self.substore).replyto = 'bar'
        prefs.touch()

    def part4SetGet(self):
        self.assertEquals(IPrefs(self.substore).replyto, 'bar')

    def part5SetGet(self):
        prefs = IPrefs(self.substore)
        IPrefs(self.substore).emergencyContact = 'baz'
        prefs.touch()

    def part6SetGet(self):
        self.assertEquals(IPrefs(self.substore).emergencyContact, 'baz')
        self.assertEquals(IPrefs(self.substore).timezone, 'foo')
    
    multitxnSetGet = [part1SetGet, part2SetGet, part3SetGet, part4SetGet, part5SetGet]


f = lambda o: IResource(IPrefs(o))

class PrefupDisplay(unittest.TestCase):
    def dbSetUp(self):
        self.avatar = self.createAvatar([Prefup])

    def set(self, attr, value):
        setattr(f(self.avatar), attr, value)

    def part1TouchNForward(self):
        self.set("timezone", "foo")
    
    def part2TouchNForward(self):
        self.assertEquals(f(self.avatar).timezone, "foo")

    def part3TouchNForward(self):
        self.set("replyto", "bar")
    
    def part4TouchNForward(self):
        self.assertEquals(f(self.avatar).replyto, "bar")

    def part5TouchNForward(self):
        self.set("emergencyContact", "baz")
    
    def part6TouchNForward(self):
        self.assertEquals(f(self.avatar).emergencyContact, "baz")

    multitxnTouchNForward = [part1TouchNForward, part2TouchNForward,
        part3TouchNForward, part4TouchNForward, part5TouchNForward,
        part6TouchNForward]
