
from twisted.trial import unittest

from quotient import stats

class SizeFormatting(unittest.TestCase):
    def _test(self, data, *a, **kw):
        for (bytes, s) in data:
            self.assertEquals(stats.formatSize(bytes, *a, **kw), s)

    def testDefaults(self):
        self._test([(0, '0.00 bytes'),
                    (1, '1.00 byte'),
                    (9, '9.00 bytes'),
                    (72, '72.00 bytes'),
                    (1023, '1023.00 bytes'),
                    (1024, '1.00 kilobyte'),
                    (1025, '1.00 kilobyte'),
                    (2048, '2.00 kilobytes'),
                    (2276, '2.22 kilobytes'),
                    (1024 * 1024, '1.00 megabyte'),
                    (1024 * 1024 * 1.5, '1.50 megabytes'),
                    (2 * 1024 ** 3, '2.00 gigabytes'),
                    (3 * 1024 ** 4, '3.00 terabytes'),
                    (4 * 1024 ** 5, '4.00 petabytes'),
                    (5 * 1024 ** 6, 'a whole lot of bytes (%d)' % (5 * 1024 ** 6,))])

    def testNonPluralized(self):
        self._test([(0, '0.00 byte'),
                    (1, '1.00 byte'),
                    (9, '9.00 byte'),
                    (72, '72.00 byte'),
                    (1023, '1023.00 byte'),
                    (1024, '1.00 kilobyte'),
                    (1025, '1.00 kilobyte'),
                    (2048, '2.00 kilobyte'),
                    (2276, '2.22 kilobyte'),
                    (1024 * 1024, '1.00 megabyte'),
                    (1024 * 1024 * 1.5, '1.50 megabyte'),
                    (2 * 1024 ** 3, '2.00 gigabyte'),
                    (3 * 1024 ** 4, '3.00 terabyte'),
                    (4 * 1024 ** 5, '4.00 petabyte'),
                    (5 * 1024 ** 6, 'a whole lot of byte (%d)' % (5 * 1024 ** 6,))],
                   pluralize=False)

    def testNoFractionalPart(self):
        self._test([(0, '0 bytes'),
                    (1, '1 byte'),
                    (9, '9 bytes'),
                    (72, '72 bytes'),
                    (1023, '1023 bytes'),
                    (1024, '1 kilobyte'),
                    (1025, '1 kilobyte'),
                    (2048, '2 kilobytes'),
                    (2276, '2 kilobytes'),
                    (1024 * 1024, '1 megabyte'),
                    (1024 * 1024 * 1.5, '2 megabytes'),
                    (2 * 1024 ** 3, '2 gigabytes'),
                    (3 * 1024 ** 4, '3 terabytes'),
                    (4 * 1024 ** 5, '4 petabytes'),
                    (5 * 1024 ** 6, 'a whole lot of bytes (%d)' % (5 * 1024 ** 6,))],
                   fractionalPart=False)

        def testNonPluralAndNoFractional(self):
            self._test([(0, '0 byte'),
                    (1, '1 byte'),
                    (9, '9 byte'),
                    (72, '72 byte'),
                    (1023, '1023 byte'),
                    (1024, '1 kilobyte'),
                    (1025, '1 kilobyte'),
                    (2048, '2 kilobyte'),
                    (2276, '2 kilobyte'),
                    (1024 * 1024, '1 megabyte'),
                    (1024 * 1024 * 1.5, '2 megabytes'),
                    (2 * 1024 ** 3, '2 gigabyte'),
                    (3 * 1024 ** 4, '3 terabyte'),
                    (4 * 1024 ** 5, '4 petabyte'),
                    (5 * 1024 ** 6, 'a whole lot of byte (%d)' % (5 * 1024 ** 6,))],
                   fractionalPart=False, pluralize=False)
