# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from atop.test import unittest

from atop import sched
from atop import credup

from twisted.cred import checkers
from twisted.internet import defer
from twisted.python import log
from twisted.internet.error import DNSLookupError

from quotient.test import helper

from quotient import sysup
from quotient import grabbers
from quotient import searchup

class SystemWideSettings(unittest.TestCase, helper.ProvisionerMixin):

    storePowerups = [credup.StoreCredentialsChecker]

    def dbSetUp(self):
        self.avatar = self.createAvatar([sysup.Adminup, grabbers.GrabberPowerup, searchup.Searchup])
        self.admin = sysup.ISystemAdministrator(self.avatar)
        self.control = sysup.IControlPanel(self.store)
        checker = checkers.ICredentialsChecker(self.store)
        checker.addDomain("localhost")
        self.sched = sched.SchedulerService(self.store)
        self.sched.startService()

    def dbTearDown(self):
        nc = searchup.ISearch(self.avatar).checkpointDefer
        nc and nc.cancel()
        self.sched.stopService()

    def dbtestUserCreation(self):
        self.control.allowAction(sysup.IControlPanel.USER_CREATION, False)
        self.control.allowAction(sysup.IControlPanel.TICKET_REDEMPTION, False)
        self.assertRaises(sysup.AdministrativeRestriction,
                          sysup.makeUser, self.store, "username", "password",
                                         (), True)
        self.assertRaises(sysup.AdministrativeRestriction,
                          sysup.makeUser, self.store, "username", "password",
                                         (), False)

        self.control.allowAction(sysup.IControlPanel.USER_CREATION, True)
        self.assertRaises(sysup.AdministrativeRestriction,
                          sysup.makeUser, self.store, "username", "password",
                                         (), False)
        avatar = sysup.makeUser(self.store, "username", "password", (), True)

        self.control.allowAction(sysup.IControlPanel.TICKET_REDEMPTION, True)
        self.control.allowAction(sysup.IControlPanel.USER_CREATION, False)
        self.assertRaises(sysup.AdministrativeRestriction,
                          sysup.makeUser, self.store, "username2", "password",
                                         (), True)
        avatar = sysup.makeUser(self.store, "username2", "password", (), False)

        self.control.allowAction(sysup.IControlPanel.USER_CREATION, True)
        avatar = sysup.makeUser(self.store, "username3", "password", (), False)
        avatar = sysup.makeUser(self.store, "username4", "password", (), True)

    def dbtestGrabberProhibition(self):
        self.control.allowAction(sysup.IControlPanel.EXTERNAL_GRABBING, False)

        gp = grabbers.IGrabberAdder(self.avatar)
        gf = gp.addPopGrabber("u", "p", "...", (1234, False), "foo@bar", 0)

        result = gf.run()
        self.assertIdentical(result, None)
        self.assertIdentical(gf.doneDeferred, None)

        self.control.allowAction(sysup.IControlPanel.EXTERNAL_GRABBING, True)

        result = gf.run()
        self.failUnless(isinstance(result, defer.Deferred))
        unittest.deferredError(result)
        e = log.flushErrors(DNSLookupError)
        self.assertEquals(len(e), 1)

    def dbtestIndexingProhibition(self):
        L = []
        su = searchup.ISearch(self.avatar)
        su.systemBlockedIndexing = lambda: L.append(None)

        self.control.allowAction(sysup.IControlPanel.MESSAGE_INDEXING, False)
        su.run()
        self.failUnless(L)

        L = []
        self.control.allowAction(sysup.IControlPanel.MESSAGE_INDEXING, True)
        su.run()
        self.failIf(L)
        del su.systemBlockedIndexing
