
from twisted.web.resource import IResource
from atop import powerup
from atop.store import Item, Pool, UIDIndex
from atop.test.unittest import TestCase
from nevow.testutil import FakeRequest
from quotient.web.pools import PoolBrowser
from quotient.web import webup
from quotient import prefup

from twisted.web.microdom import parseString
from twisted.web.domhelpers import findElementsWithAttribute as findwa
from twisted.web.domhelpers import findNodes as findn
from twisted.web.domhelpers import findNodesNamed as findnn
from twisted.web.domhelpers import gatherTextNodes as gather

import webbrowser
import os

def nextargs(dom):
    for node in findnn(dom,'a'):
        if len(node.childNodes) == 1:
            if gather(node) == 'Next':
                text = node.getAttribute('href')
                d = dict([x.split('=') for x in text.split('?')[-1].split('&')])
                return d.get('index'), d.get('start'), d.get('startid')

OPEN_BROWSER = False

class TabularDataBrowserTest(TestCase):

    def requestFromBrowser(self, index=None, start=None, startid=None, openbrowser=OPEN_BROWSER):
        if index:
            argh = {'index': [index],
                    'direction': ['forewards']}
        else:
            argh = {}
        if start is not None:
            argh['start'] = [start]
        if startid is not None:
            argh['startid'] = [startid]
        dr = FakeRequest(args=argh, avatar=IResource(self.bob))
        PoolBrowser(self.p).render(dr)
        if openbrowser:
            p = os.path.abspath(self.mktemp()+'.html')
            f = open(p,'wb')
            f.write(dr.v)
            f.flush(); f.close()
            u = 'file://' + p
            webbrowser.open(u)
        self.failUnless(dr.finished)
        dom = parseString(dr.v)
        return dom

    def firstKeyFromIndexQuery(self, dom, index):
        flist = findwa(dom, "class", "freeform-list-item")[0]
        storeID = int(findnn(flist, 'a')[0].getAttribute('href').split('/')[-1].split('?')[0])
        item = self.bob.getItemByID(storeID)
        return item.getIndexKey(self.p, index)

    def dbSetUp(self):
        self.bob = self.createAvatar([powerup.MetaPowerup, prefup.Prefup, webup.WebPowerup])
        self.p = Pool(self.bob)
        a = UIDIndex(); a.name='a'
        b = UIDIndex(); b.name='b'
        self.p.addIndex(a)
        self.p.addIndex(b)
        for x in range(100):
            i = Item(self.bob, str(x))
            i.setIndexKey("a", x)
            i.setIndexKey("b", 0)
            self.p.addItem(i)
        # SANITY CHECK
        ol = list(self.p.queryIndex('b'))
        self.assertEquals(len(ol), 100)
        self.assertEquals(len(list(self.p.queryIndex('b', startKey=0, startItemID=ol[50].storeID))), 50)

    def dbtestDefaultPaging(self):
        dom = self.requestFromBrowser()
        key = self.firstKeyFromIndexQuery(dom, 'uid')
        self.assertEquals(key, 99)
        self.p.suggestIndex('uid', False)
        dom = self.requestFromBrowser()
        key = self.firstKeyFromIndexQuery(dom, 'uid')
        self.assertEquals(key, 0)
    dbtestDefaultPaging.skip = "This test suite doesn't wait for deferreds, so this fails"

    def dbtestPaging(self):
        dom = self.requestFromBrowser('a')
        key = self.firstKeyFromIndexQuery(dom, 'a')
        self.assertEquals(key, 0)
        dom = self.requestFromBrowser('a', '25')
        key = self.firstKeyFromIndexQuery(dom, 'a')
        self.assertEquals(key, 25)
    dbtestPaging.skip = "This test suite doesn't wait for deferreds, so this fails"

    def dbtestCommonIndexKey(self):
        dom = self.requestFromBrowser('b')
        key = self.firstKeyFromIndexQuery(dom, 'a')
        aindex, astart, astartid = nextargs(dom)
        self.assertEquals(key, 0)
        self.assertEquals(aindex, 'b')
        dom2 = self.requestFromBrowser(aindex, astart, astartid)
        key2 = self.firstKeyFromIndexQuery(dom2, 'a')
        self.assertEquals(key2, 25)
    dbtestCommonIndexKey.skip = "I have no idea what this is testing, but it's wrong"
