# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

from quotient.proto.twisted_sip import IContact, ILocator
from quotient.proto import sexp
from quotient.proto.astmanager import AstManager
from quotient.web import root
from quotient.serveup import IServer
from quotient.application import IVoiceApp, VoiceApp
from quotient.provisioner import EPermission, ProvisionedPowerup
from quotient import taglib, timezones, prefup

from atop.powerup import Powerup, Storeup
from atop.store import Item
from atop import sched, credup, tpython

from nevow import tags
import formless
from formless import webform
from twisted.internet import reactor, protocol
from twisted.application import strports
from twisted.python import log, components
from twisted.cred.error import UnauthorizedLogin
from twisted.persisted.styles import Versioned
from twisted.runner import procutils
from twisted.spread import pb

import os, time, datetime

class IAsteriskOverlord(components.Interface):
    "We control the horizontal and the vertical."

class VoiceStoreup(Storeup):

    persistenceForgets = ['manager']
    manager = None
    
    def setUpPools(self, store):
        store.setComponent(IAsteriskOverlord, self)
        self.activate()

    def placeCall(self, username, dialstring):
        if not self.manager or not self.manager.transport.connected:
            self.setupManagerAndCall(username, dialstring)
        else:
            try:
                self.manager.originate(username, dialstring, 'generic-customer1')
            except AttributeError: # hurrah, fault-tolerant computing
                self.setupManagerAndCall(username, dialstring)

    def setupManagerAndCall(self, username, dialstring):
        f = protocol.ClientCreator(reactor, AstManager)
        f.connectTCP('127.0.0.1',5038).addCallback(self._astLogin
            ).addCallback(lambda m: m.originate(username, dialstring, 'generic-customer1')
            ).addErrback(log.msg)

    def _astLogin(self, m):
        self.manager = m
        m.login('manager','tonicpencil')
        return m

    def activate(self):
        pass

class PhoneAddress(formless.String):

    def coerce(self, val):
        if '@' in val:
            if not val.startswith("sip:"):
                val = "sip:" + val
            return val
        num = ''.join([c for c in val if c.isdigit()])
        if len(num) == 10:
            return "1"+num
        elif len(num) == 11 and num[0] == '1':
            return num
        else:
            raise formless.InputError("Please enter a SIP URL or a North American ten-digit phone number.")

class IVoiceup(formless.TypedInterface):
    altcontact = PhoneAddress(description="SIP URL or North American phone number to forward calls to when you're offline")
    dollars= formless.Real(description="How much money you have for phone calls.")
    def placeCall(self,dialstring=formless.String(description='Phone number or SIP URL')):
        "Place a call"
        pass
    placeCall = formless.autocallable(placeCall)

    def timedCall(self,hour=formless.Choice(choices=range(1,13)),
                  minute=formless.Choice(choice=range(0,65,5)),
                  ampm=formless.Choice(choicees=('am','pm'))):
        pass


class VoiceupDisplay(root.Base):
    def render_online(self, context, data):
        return self.original.isOnline

    document = taglib.page(title="Voice")[
        tags.p["Registered: ", render_online],
        webform.renderForms(),
    ]


class DefaultPlan:

    rate = 0.0005 # dollars/second

    def rateFor(self, extension):

        if extension.startswith('1'):
            return self.rate
        else:
            raise LookupError, "no rate available"



class Voiceup(ProvisionedPowerup, Versioned):
    __implements__ = IContact, ILocator, IVoiceup, ProvisionedPowerup.__implements__
    storeDependencies = [VoiceStoreup]
    dependencies = [VoiceApp]

    isOnline = False
    phoneID = None
    phonePIN = None
    altcontact = None
    powerupName = "Phone"
    dollars = 0
    plan = DefaultPlan()

    persistenceVersion = 2

    def upgradeToVersion2(self):
        self.touch()
        self.name = self.powerupName = "Phone"

    def upgradeToVersion1(self):
        self.touch()
        self.provisioner.touch()
        self.provisioner.permissioner.addPermission(EPermission.SIP_LOGIN, 12)

    def setUpPools(self, avatar):
        # ProvisionedPowerup.setUpPools(self, avatar)
        self.invisible = False
        avatar.setComponent(IContact, self)
        self.phoneID = avatar.storeID

    def online(self):
        self.isOnline = True
    def offline(self):
        self.isOnline = False

    def recordIncomingCall(self, name, uri):
        self.store.transact(lambda: IVoiceApp(self.store).appRoot.addItem(Call(self.store, name, uri, "from", time.gmtime())))

    def recordOutgoingCall(self, name, uri):
        self.store.transact(lambda: IVoiceApp(self.store).appRoot.addItem(Call(self.store, name, uri, "to", time.gmtime())))

    def placeCall(self, dialstring):
        if self.isOnline:
            def _():
                for d in self.store.iterpools():
                    if isinstance(d, credup.Domain):
                        return d.name
            domain = self.store.transact(_)
            me = "sip:%s@%s" %(self.store.name, domain)
        elif self.altcontact:
            me = self.altcontact
        else:
            return "You're not registered and you haven't specified an alternate contact."
        IAsteriskOverlord(self.store.store).placeCall(me, dialstring)

    def timedCall(self, hours, minutes, ampm):
        #self.placeCall("timedcall."+self.store.name)
        rm = RecordMessage()
        call = TimedCall(self.store)


    def availableSecondsFor(self, extension):
        try:
            return self.dollars / self.plan.rateFor(extension)
        except LookupError:
            return -1

    def expend(self, seconds, ext):
        amt = self.plan.rateFor(ext) * seconds
        log.msg("VOICE EXPEND: $%s - %s, %s secs to %s" % (amt, self.store.name, seconds, ext))
        self.dollars -= amt
        self.touch()

class Accounter(sexp.SymbolicExpressionReceiver):
    # > (query available-seconds fred 13018137829)
    # < (available-seconds 7491)

    def __init__(self, store):
        self.store = store
        sexp.SymbolicExpressionReceiver.__init__(self)

    def symbolicExpressionReceived(self, expr):
        meth = getattr(self,"do_"+expr[0],None)
        if meth:
            try:
                meth(expr[1:])
            except Exception, e:
                self.sendSymbolicExpression(("error", e.__class__.__name__, str(e)))
        else:
            self.sendSymbolicExpression(('error', "No such method"))

    def do_query(self, query):
        username, extension = query[1:3]
        try:
            ava = credup.preauth(self.store, username)
        except UnauthorizedLogin:
            ava = None
            log.msg("Unknown VoIP user " + username)

        if not ava:
            self.sendAvailableSeconds(-1)
        else:
            v = IContact(ava)
            self.sendAvailableSeconds(v.availableSecondsFor(extension))

    def sendAvailableSeconds(self, sec):
        self.sendSymbolicExpression(("available-seconds",sec))

class AccounterFactory(protocol.Factory):

    def __init__(self, store):
        self.store = store

    def buildProtocol(self, addr):
        p = Accounter(self.store)
        p.factory = self
        return p

class Call(Item):

    def __init__(self, store, name, uri, kind, time):
        self.uri = uri
        self.time = time
        self.kind = kind
        Item.__init__(self, store, name)


class TabularCall(root.TabularBase):

    def render_kind(self, context, data):
        return data.kind
    root.column(render_kind, True)

    def render_name(self, context, data):
        return data.name
    root.column(render_name, True)

    def render_uri(self, context, data):
        return data.uri.toString()
    root.column(render_uri, True)

    def render_time(self, context, data):
        # convert to datetime object
        local = tpython.mkgmtime(data.time)
        dtobj = datetime.datetime.fromtimestamp(local, timezones.Local)
        # and here's where we have to respect the user's account setting, hoorj
        astz = dtobj.astimezone(prefup.IPrefs(data.store).timezone)
        return astz.strftime('%Y/%m/%d %H:%M:%S')
    root.column(render_time, True)

class TimedCall(Item, sched.Schedulable):
    pass

class UnlimitedFreePlan:
    def rateFor(self, extension):
        return 0


