import os
import time
import calendar
calendar.setfirstweekday(calendar.MONDAY)

from datetime import date

from nevow.tags import *
from nevow.stan import slot
from nevow.url import here
from nevow import liveevil
from nevow import inevow

from quotient.taglib import *
from quotient.web import root
from quotient.web import pools
from quotient.cleanup import IClean
from quotient.mimemessage import EmailPool
from quotient import application

from quotient import web

from twisted.python import util

billboards = util.sibpath(web.__file__, 'billboards')

class AppRenderer(root.Base):
    def render_billboard(self, ctx, data):
        flname = os.path.join(billboards, '%s.html' % self.original.name)
        if os.path.exists(flname):
            return open(flname).read()
        return self.original.name, " is coming soon."

    document = page(title="Application")[
        ## Here we load the Billboard page for this application
        render_billboard
    ]

    def renderHTTP(self, request):
        appNav = list(self.original.getApplicationNav())
        if appNav:
            redirectTo = list(self.original.getApplicationNav())[0]
            request.redirect(str(request.URLPath().sibling(str(redirectTo.storeID))))
            return "<html>Redirecting</html>"
        return root.Base.renderHTTP(self, request)
    


class DivmodRenderer(root.Base):
    document = page(title="Application")[
        "nothing"]


def divmodFactory(requestCtx):
    """Given a RequestContext, produce the instance of DivmodApplication
    installed on the avatar. This enables web code to say ctx.locate(IDivmodApp)
    """
    req = requestCtx.tag
    return application.IDivmodApp(
        getattr(req.getSession().getLoggedInRoot(),'store',None))


class Chatola(object):
    def __init__(self):
        self.clients = []

    def join(self, client):
        client.userId = "Anonymous %s" % (len(self.clients) + 1, )
        self.broadcast(client, "%s has joined." % client.userId)
        self.clients.append(client)

    def broadcast(self, user, message):
        """Broadcast a message to every client which is connected to this realm.
        """
        script = liveevil.callJS("writeContent", "<div>%s: %s</div>" % (user.userId, message))
        for C in self.clients:
            C.sendScript(script)
            if C is user:
                C.sendScript("focusInput();")


chat = Chatola()


class ChatRenderer(root.Base):
    def render_content(self, context, data):
        """When a user enters the chat we'll render a list of the people that are currently in it
        """
        return context.tag[
            "Userlist: ",
            [(user.userId, " ") for user in chat.clients]
        ]

    def render_input(self, context, data):
        """Render an input form; a text box and a submit button. On the form, we add an onsubmit
        event handler; we pass a python callable wrapped by the liveevil.handler function which will
        be called on the server when the client side onsubmit handler is called. liveevil.handler
        automatically returns false in  the browser handler so the normal page submit does not
        occur. Instead, the second argument to handler, a string, is evaluated as javascript in the
        browser context and the result is sent to the server-side onSubmit method.
        """
        return form(
            onsubmit=liveevil.handler(self.onSubmit, "getValue('inputline');")
        )[
            input(type="text", id="inputline"), 
            input(type="submit", value="say")
        ]

    def onSubmit(self, client, text):
        """When the onSubmit method is called on the server in response to the client-side
        onsubmit javascript handler, pass the text which was in the user's input box to the
        realm's broadcast method. Broadcast will format the message with the user's nick
        and send javascript to every connected client (including this user) which will append
        this message to the "content" div so all users can see the new chat text.
        """
        chat.broadcast(client, text)

    def render_nick(self, context, data):
        return form(
            onsubmit=liveevil.handler(self.onNickChange, "getValue('nick');")
        )[
            input(type="text", id="nick", value=self.client.userId),
            input(type="submit", value="change nick"),
            span(id="nickChangeNotification")
        ]

    def onNickChange(self, client, newval):
        chat.broadcast(client, "%s is now known as %s." % (client.userId, newval))
        client.userId = newval
        client.sendScript(liveevil.callJS('nickChanged', newval))

    def render_glue(self, context, data):
        """Insert the liveevil glue necessary for the input and output conduits to be present,
        and thus serverToClient and clientToServer events to be passed.
        """
        return liveevil.glue

    def render_joinChat(self, ctx, data):
        client = liveevil.ILiveEvil(ctx.locate(inevow.IRequest).getSession())
        chat.join(client)
        self.client = client
        return ''

    document = page(title="Chat!!1")[
        render_joinChat,
        script(language="javascript")[xml("""
function writeContent(content) {
    document.getElementById('content').innerHTML += content;
}

function focusInput() {
	var il = document.getElementById('inputline');
	il.focus();
	il.value = '';
	document.getElementById('bottom').scrollIntoView();
}

function nickChanged(newNick) {
	var notification = document.getElementById('nickChangeNotification');
	notification.innerHTML = 'Your nick is now ' + newNick;
	window.setTimeout("document.getElementById('nickChangeNotification').innerHTML = '';", 5000);
	focusInput();
}

function getValue(fieldId) {
	return document.getElementById(fieldId).value;
}
""")],
        div(style="margin: 1em")[
            div(id='content', render=render_content),
            form(render=render_input),
            form(render=render_nick),
            span(render=render_glue),
            span(id='bottom')]]


def data_currentMonth(context, data):
    curtime = time.localtime()
    year = int(context.arg('year', curtime[0]))
    month = int(context.arg('month', curtime[1]))
    return year, month, calendar.monthcalendar(year, month)


class CalendarRenderer(root.Base):
    def render_month(self, context, data):
        year, month, weeksAndDays = data

        weekPattern = context.patternGenerator('calendarWeek')
        dayPattern = context.with(weekPattern()).patternGenerator('calendarDay')
        curtime = time.localtime()
        curYear, curMonth, curDate = curtime[0], curtime[1], curtime[2]

        self.pool = IClean(self.original.store).cleanPool

        calendarBody = []
        for week in weeksAndDays:
            currentWeek = weekPattern().clear()
            calendarBody.append(currentWeek)
            for day in week:
                today = None
                if curYear == year and curMonth == month and curDate == day:
                    today = 'today'
                currentDay = dayPattern(
                    id=today,
                    render=self.render_day,
                    data=(year, month, day)
                )
                currentWeek.children.append(currentDay)

        context.fillSlots('label', "%s %s" % (calendar.month_name[month], year))
        context.fillSlots('calendarBody', calendarBody)
        return context.tag

    def render_day(self, ctx, data):
        year, month, day = data
        for events in self.pool.queryIndex(
            'received', 
            startKey=(year, month, day, 0, 0, 0, 0, 0, 0), 
            endKey=(year,month,day+1,0,0,0,0,0,0)):
            events = True
            break
        else:
            events = False
        if day == 0:
            day = ''
        elif events:
            day = a(href=here.child(str(year)).child(str(month)).child(str(day)))[day]
        return ctx.tag.clear()[day]

    def render_monthSelector(self, ctx, data):
        curtime = time.localtime()
        curyear = int(ctx.arg('year', curtime[0]))
        curmonth = int(ctx.arg('month', curtime[1]))
        return form(action="", method="GET")[
            select(name='month', onchange='submit()', style="width: 1in;")[[
                option(value=month, selected=month==curmonth and 'selected' or None)[month]
                for month in range(1, 13)
            ]],
            select(name='year', onchange='submit()', style="width: 1in;")[[
                option(value=year, selected=year==curyear and 'selected' or None)[year]
                for year in range(2003, 2006)
            ]],
        ]

    document = page(title="Calendar")[
        table(data=data_currentMonth, render=render_month, style="width: 50%; margin: 1em;")[
            thead[
                tr[td(colspan=7)[slot('label')]],
                tr[[td[x] for x in calendar.weekheader(10).split()]]],
            tbody[
                tr(pattern='calendarWeek', render=lambda ctx, data: '')[
                    td(pattern='calendarDay')['day']],
                slot('calendarBody')]],
        div(render=render_monthSelector)]

    def locateChild(self, request, childSegments):
        if childSegments == ('',):
            ## We are looking for the root of the site, aka /
            return self, ()
        try:
            year, month, date = map(int, childSegments[:3])
            return DayDetail(self.original, year, month, date), childSegments[3:]
        except ValueError:
            # If the url doesn't consist of a tuple of year, month, day, or the segments
            # are not integers, then we render a 404 page.
            return rend.NotFound


class DayDetail(root.Base):
    def __init__(self, original, year, month, date):
        root.Base.__init__(self, original)
        self.year = year
        self.month = month
        self.date = date

    def render_detail(self, ctx, data):
        year = self.year
        month = self.month
        date = self.date
        return [
            p(style="margin: 15px")[a(href=here.click('../..')(year=year, month=month))['back']], 
            h1(style="margin: 15px")[year, ' ', month, ' ', date],
            pools.PoolBrowser(list(
                IClean(self.original.store).cleanPool.queryIndex(
                    'received', 
                    startKey=(year, month, date, 0, 0, 0, 0, 0, 0), 
                    endKey=(year,month,date+1,0,0,0,0,0,0))))]

    document = page(title="day")[
        render_detail]

