# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import string

from quotient.web.root import Base, TabularBase, column, action_delete
from quotient import taglib
from quotient.addressbook import IAddressBook
from quotient.smarts.person import Person, IPersonCollector
from quotient.web import pools

from nevow import inevow
from nevow import tags
from nevow import entities
import formless
from formless import webform
from formless import iformless
from formless import configurable
from nevow import blocks
from nevow.url import here
from nevow import compy


class PersonRow(TabularBase):
    def render_emailAddress(self, context, data):
        if data.emails.values():
            return data.emails.values()[0]
        else:
            return ''
    column(render_emailAddress, True, 250, "Email")

    action_delete = action_delete

    def action_addToAddressBook(self, pool):
        addPool = IAddressBook(pool.store).addressPool
        if self.original not in addPool:
            addPool.addItem(self.original)

    def action_removeFromAddressBook(self, pool):
        addPool = IAddressBook(pool.store).addressPool
        if self.original in addPool:
            addPool.removeItem(self.original)


personName = lambda context, data: taglib.headerToUTF8(data.name or '<No Name>')
personFormSkin = tags.form(pattern='freeform-form', id=tags.slot('form-id'), action=tags.slot('form-action'), method='POST', _class="edit-person-form")[
    ## The pattern which all fields will use
    tags.invisible(pattern='argument', render=lambda c, d: '')[
        blocks.block(width="100px")[tags.slot('label')], 
        blocks.block(width="175px")[tags.slot('input')]],
    blocks.line()[
        blocks.block(width="290px")[
            tags.slot('argument!!firstName')],
        blocks.block(width="290px")[
            tags.slot('argument!!lastName')]],
    blocks.block(width="290px")[
        tags.slot('argument!!address')[
            tags.invisible(pattern="argument!!address")[
                blocks.line()["Address"],
                tags.slot('input')]]],
    blocks.block(width="290px")[
        blocks.line()[tags.slot('argument!!phone')],
        blocks.line()[tags.slot('argument!!homePhone')],
        blocks.line()[tags.slot('argument!!businessPhone')],
        blocks.line()[tags.slot('argument!!cell')],
        blocks.line()[tags.slot('argument!!homeFax')],
        blocks.line()[tags.slot('argument!!email')],
        blocks.line()[tags.slot('argument!!homeEmail')],
        blocks.line()[tags.slot('argument!!businessEmail')]],
    blocks.line()[
        blocks.block(width="400px")[entities.nbsp],
        tags.slot('form-button')]]


class PersonDisplay(Base):
    def render_personSummary(self, ctx, data):
        from quotient.proto import twisted_sip
        from atop import powerup
        from quotient.proto.smtpout import OutgoingMailup

        emailer = powerup.IPowerStation(data.store).installPowerup(OutgoingMailup).storeID
        dialer = twisted_sip.IContact(data.store).storeID

        from quotient.smarts import phone
        return tags.ul(_class="contact-info")[
            [tags.li[
                tags.a(href="%s/freeform_post!!placeCall?dialstring=%s" % (dialer, V))[
                    tags.img(src="/images/badges/phone.png"),
                    tags.strong[K == '' and 'phone' or K],
                    entities.nbsp,
                    V]]
            for K, V in data.phones.iteritems() if V],
            [tags.li[
                tags.a(href="%s?newMessageTo=%s" % (emailer, V))[
                    tags.img(src="/images/badges/email.png"), 
                    tags.strong[K == '' and 'email' or K], 
                    entities.nbsp,
                    V]]
            for K, V in data.emails.iteritems()if V],
            (not bool(getattr(data, 'address', None))) and ' ' or tags.span(style="white-space: nowrap")[
                tags.img(src="/images/badges/address.png"),
                tags.strong["address"], 
                entities.nbsp,
                data.address]]

    document = taglib.page(title=("Person Detail: ", personName))[
        tags.div(style="border-bottom: 1px dotted black;")[
            tags.h1[
                personName],
            tags.div(style="border-bottom: 1px dotted black; padding: 5px;")[
                render_personSummary],
            tags.div(style="border-bottom: 1px dotted black; padding-top: 5px; padding-bottom: 5px;")[
                blocks.collapser(
                    'Show Recent Messages',
                    'Hide Recent Messages',
                    webform.renderForms('recentMessages'),
                    False)],
            tags.div(style="padding-top: 5px; padding-bottom: 5px;")[
                blocks.collapser(
                    'Show Edit Person',
                    'Hide Edit Person',
                    webform.renderForms('editPerson')[
                        personFormSkin],
                    True)]]]

    def configurable_editPerson(self, ctx):
        person = self.original
        ## This is SO just a temporary hack and should go away asap
        D = ctx.locate(iformless.IFormDefaults).getAllDefaults('editPerson')
        D['firstName'] = person.firstName
        D['lastName'] = person.lastName
        D['address'] = getattr(person, 'address', '')
        D['phone'] = person.phones.get('', '')
        D['homePhone'] = person.phones.get('home phone', '')
        D['businessPhone'] = person.phones.get('business phone', '')
        D['cell'] = person.phones.get('cell', '')
        D['homeFax'] = person.phones.get('home fax', '')
        D['businessFax'] = person.phones.get('business fax', '')
        D['email'] = person.emails.get('', '')
        D['homeEmail'] = person.emails.get('home email', '')
        D['businessEmail'] = person.emails.get('business email', '')

        return PersonEditor(person)

    def configurable_recentMessages(self, ctx):
        rv = pools.PoolBrowser(list(self.original.recentMessages(20)))
        ctx.remember(rv, inevow.IRendererFactory)
        return rv

    def render_gallery(self, ctx, data):
        if IThumbnail(self.original.store, default=None) is None:
            return []
        return [tags.h4['Gallery'],
                webform.renderForms('gallery')]

    def configurable_gallery(self, ctx):
        person = self.original
        from quotient.smarts.image import IThumbnail
        ic = IThumbnail(person.store)
        return pools.PoolBrowser(list(ic.extractsFrom(person)), suggestedView="Gallery")

    def getDynamicChild(self, name, request):
        from quotient.web import root
        try:
            i = int(name)
        except ValueError:
            return


class AddressBookDisplay(Base):
    def render_alphabet(self, context, data):
        start = context.arg('start', ' ')[0]
        alpha = ['ABC', 'DEF', 'GHI', 'JKL', 'MNO', 'PQR', 'STUV', 'WXYZ']
        return tags.ol(_class="alphabet")[
            tags.a(href=here.add('direction', 'forewards').add('index', 'lastName'))[
            tags.li(id=start == ' ' and 'address-tab-selected' or None)[
                'All']], 
            [tags.a(href=here.add('start', label[0]).add('direction', 'forewards').add('index', 'lastName'))[
                tags.li(id=label[0] == start and 'address-tab-selected' or None)[
                    label]]
            for label in alpha]]

    def configurable_addressBookPool(self, ctx):
        rv = pools.PoolBrowser(self.original.addressPool)
        ctx.remember(rv, inevow.IRendererFactory)
        return rv

    def configurable_addPerson(self, ctx):
        return PersonAdder(self.original)

    document = taglib.page(title="Address Book")[
        render_alphabet,
        tags.h1["Address Book"],
        webform.renderForms('addressBookPool'),
        blocks.collapser(
            'Show Add Person',
            'Hide Add Person',
            webform.renderForms('addPerson')[
                personFormSkin])]



personSchema = [
    formless.Argument('firstName', formless.String()),
    formless.Argument('lastName', formless.String()),
    formless.Argument('address', formless.Text()),
    formless.Argument('phone', formless.String()),
    formless.Argument('homePhone', formless.String()),
    formless.Argument('businessPhone', formless.String()),
    formless.Argument('cell', formless.String()),
    formless.Argument('homeFax', formless.String()),
    formless.Argument('businessFax', formless.String()),
    formless.Argument('email', formless.String()),
    formless.Argument('homeEmail', formless.String()),
    formless.Argument('businessEmail', formless.String())]


class PersonEditor(configurable.Configurable):
    def getBindingNames(self, *args):
        return ['editPerson']

    def bind_editPerson(self, ctx):
        return formless.MethodBinding(
            'editPerson',
            formless.Method(arguments=personSchema))

    def editPerson(self, firstName, lastName, address, phone, homePhone, businessPhone, cell, homeFax, businessFax, email, homeEmail, businessEmail):
        person = self.original
        person.touch()
        person.setName(firstName or '', lastName or '')
        person.address = address or ''
        person.setPhoneNumber(phone or '')
        person.setPhoneNumber(homePhone or '', 'home phone')
        person.setPhoneNumber(businessPhone or '', 'business phone')
        person.setPhoneNumber(cell or '', 'cell')
        person.setPhoneNumber(homeFax or '', 'home fax')
        person.setPhoneNumber(businessFax or '', 'business fax')
        person.setEmailAddress(email or '')
        person.setEmailAddress(homeEmail or '', 'home email')
        person.setEmailAddress(businessEmail or '', 'business email')


class PersonAdder(PersonEditor):
    def getBindingNames(self, *args):
        return ['newPerson']

    def bind_newPerson(self, ctx):
        return formless.MethodBinding(
            'newPerson',
            formless.Method(arguments=personSchema))

    def newPerson(self, firstName, lastName, **kw):
        ## This code is horrible.
        firstName = firstName or ''
        lastName = lastName or ''
        ab = self.original
        ava = ab.store
        collector = IPersonCollector(ava)
        newPersonName = firstName + ' ' + lastName
        p = collector.newPerson(newPersonName)
        ab.addPerson(p)
        self.original = p
        self.editPerson(firstName, lastName, **kw)
        return p

