# -*- test-case-name: "quotient.test.test_webui" -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.


import itertools
import os
import random
from time import time as now
import datetime
import sre

from atop import tpython
from atop import credup

from twisted.web import server, static
from twisted.python import util
from twisted.python import reflect
from twisted.python import components
from twisted.python import log
from twisted.cred.checkers import ICredentialsChecker

from quotient import web
from quotient.web import webinterfaces
from quotient import taglib
from quotient.iquotient import ITabularRenderer
from quotient import timezones

import nevow
import formless
from formless import iformless, webform
from nevow import inevow
from nevow import appserver
from nevow import blocks
from nevow import rend
from nevow import loaders
from nevow import stan
from nevow.tags import *
from nevow import guard
from nevow import url


templatePath = util.sibpath(web.__file__, 'templates')

class RootCrudMixin:
    """I require ChildPrefixMixin. ONLY MIX ME IN IF YOU ARE REALLY THE ROOT, I
    WILL GO AWAY LATER.
    """

    def child_images(self, request):
        return taglib.images(self)

    def child_ximages(self, request):
        try:
            import exponent
        except:
            return rend.NotFound
        return static.File(util.sibpath(exponent.__file__, 'images'))

    def child_css(self, request):
        return taglib.css(self)

    child_freeform_css = static.File(util.sibpath(formless.__file__, 'freeform-default.css'))
    child_scripts = static.File(util.sibpath(web.__file__, 'scripts.js'), 'application/x-javascript')

    def child_reset(self, request):
        from atop import credup
        from quotient import reset
        a = credup.preauth(self.original.store, "janitor")
        return inevow.IResource(reset.IPasswordResetter(a))

    child_mozbinding = blocks.blocks_child

class Base(rend.Page):
    adaptsAvatar = False
    """If a base subclass adapts the avatar to IResource, adaptsAvatar should
    be true and the .avatar attribute will be set to the Avatar. If a Base subclass
    adapts an object which is not the avatar (such as a Pool, or a Powerup or an
    Item), it should set adaptsAvatar to False, and .avatar will be set to
    original.store.
    """
    def __init__(self, original):
        self.original = original
        if self.adaptsAvatar:
            self.avatar = original
        else:
            self.avatar = original.store
        rend.Page.__init__(self, docFactory = loaders.stan(self.document))
        self.remember(taglib.DefaultLinkGenerator(), taglib.ILinkGenerator)
        self.remember(self.avatar, credup.IAvatar)

    def renderHTTP(self, request):
        request.setHeader("Content-type", 'text/html; charset=UTF-8')

        return self.avatar.transact(rend.Page.renderHTTP, self, request)

    def webFormPost(self, request, res, configurable, ctx, bindingName, args):
        avatar = ctx.locate(credup.IAvatar)
        try:
            return avatar.transact(rend.Page.webFormPost, self, request, res, configurable, ctx, bindingName, args)
        except formless.InputError:
            ## Hack: try calling ICommander.postForm, then call ICommander.command
            from quotient import commander
            if not args.has_key('input'):
                raise
            inp = args['input'][0]
            ## DOUBLE HACK! pass charset along when doing funky redirect
            charset = args['_charset_'][0]
            cmd = commander.ICommand(res.avatar)
            result = res.avatar.transact(cmd.commandRaw, request, inp, charset)
            return res

def render_url(context, data):
    req = context.locate(inevow.IRequest)
    return req.path


class Quotient404(Base):
    def __init__(self, original, developer=False):
        self.developer = developer
        Base.__init__(self, original)

    def render_devinfo(self, context, data):
        if self.developer:
            R = context.locate(inevow.IRequest)
            return "Resources involved in lookup:", ul[[
                li[str(res)]
                for res in R.resources
            ]]
        return "Try these links:", ul[
            li[a(href="javascript:history.back(1)")["Back"]],
            li[a(href="/me/")["Home"]]
        ]

    document = taglib.page(title="Not found")[
    h1["404 Not Found"],
    p[
        "Sorry, the object located at the url ",
        render_url,
        " was not found. "
    ],
    render_devinfo
]


pat = sre.compile(' \w+::')


class CommandLineGenerator(components.Adapter):
    """command format

    /methodName arg kw::hello kw2::world
    """
    def getResourceForRequest(self, request):
        if request.postpath[-1] == 'quotient_command_line':
            userinput = request.args['input'][0]
            if userinput.startswith('/'):
                #del request.args['input']
                userinput[1:].split()[0]
                methodName = userinput[1:].split()[0]
                request.postpath[-1] = 'freeform_post:%s' % methodName

                rest = userinput[len(methodName) + 1:]
                if not rest:
                    return
                ## rest is a set of keyword arguments deliniated by ::
                ## eg key::value can have spaces anotherKey:: anotherValue (note space between colons and value)
                ## -> {'key': ['value can have spaces'], 'anotherKey': [' anotherValue (not space between colons and value)']}
                ## question: should we lstrip the value to get rid of the space? or strip the value so that spaces don't matter?

                ## '----' represents the direct object, the one argument which can be to the right
                ## of the method name and doesn't have to be a keyword argument (although it can be)
                end = None
                for match in pat.finditer(rest):
                    prevend = end
                    start, end = match.start(), match.end()
                    if prevend is None:
                        do = rest[:start]
                        if do:
                            request.args['----'] = [do]
                    else:
                        request.args.setdefault(arg, []).append(rest[prevend:start])
                    arg = rest[start+1:end-2]
                if end is None:
                    request.args['----'] = [rest]
                else:
                    request.args.setdefault(arg, []).append(rest[end:])
                ## Now we have mutated the request so that the freeform form post parser will handle it!
            else:
                request.postpath[-1] = 'freeform_post:*' # it is impossible to have a nice day


class AtopSite(appserver.NevowSite):
    generatorFactories = [
        CommandLineGenerator,
        appserver.PrettyPrinter,
    ]


class IUserSignup(formless.TypedInterface):
    def signup(self,
        request=formless.Request(),
        username=formless.String(),
        password=formless.Password(),
        realname=formless.String()):
        """Create account

        Create a new account on this server.
        """
        pass
    signup = formless.autocallable(signup)


class LoginMixin:
    def render_loginAction(self, context, data):
        req = context.locate(inevow.IRequest)
        rest = '/'.join(req.prepath)
        if not rest or not rest.startswith('me'):
            rest = 'me/'
        return form(id='signin', method='POST', action='/%s/%s' % (guard.LOGIN_AVATAR, rest))[
            dl(id="login-form")[
                dt["Username"],
                dd[input(id='username', type='text',name='username')],
                dt["Password"],
                dd[input(id='password', type='password',name='password')]],
            input(id='login', type='submit',value='go')]

class LoginOnly(Base, LoginMixin):
    def render_url(self, context, data):
        return context.locate(inevow.IRequest).path

    document = (
    xml('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n'),
    html(xmlns="http://www.w3.org/1999/xhtml", lang="en", **{'xml:lang': "en"})[
        head[
            title["Error: Please log in"],
            link(rel="stylesheet", href="/css"),
        ],
        body[
            div(style="border: 1px dotted black; margin: 1em auto; height: 240px; width: 320px; padding: 2em")[
                p["Please log in to access '", render_url, "':"],
                invisible(render=directive('loginAction'))
            ]
        ]
    ]
)

import os.path
from twisted.python.util import sibpath

sel = "/images/anonymous-tabs-highlighted/%s.png"
notsel = "/images/anonymous-tabs/%s.png"

try:
    ## Switch depending on whether we are running the actual divmod.com instance,
    ## Or somebody's personal copy of quotient.
    import exponent
    footer = ul[
        li[a(href="About")["About Divmod"]],
        li[a(href="Privacy")["Privacy Policy"]],
        # li[a(href="Terms")["Terms of Use"]],
        ]
    billboards = ["Mail", "Voice", "Chat", "People", "Album", "Calendar", "Todo", "Find", "Settings"]
    anonymous_logo = "/ximages/anonymous_logo.png"
    fileDir = os.path.join(os.path.split(exponent.__file__)[0], 'billboards')
    announce_sidebar = div(id="sidebar")[
        img(src="/images/sidebar/dyk-top.png"),
        div(id="sidebar-content")[
            form(
                action="http://divmod.org/users/mailman.twistd/subscribe/divmod.com-announce", 
                method="POST",
                style="margin: 0px; margin-right: 5px;")[
                p["""divmod.com is currently in alpha testing. Subscribe to the divmod announce mailing list to be 
                    notified when signup is enabled."""],
                input(type="text", name="email", value="Your email address", style="width: 100%"),
                input(type="hidden", name="fullname", value=""),
                input(type="hidden", name="pw", value=""),
                input(type="hidden", name="pw-conf", value=""),
                p[
                    input(type="submit", name="subscribe", value="Subscribe")]
            ]
        ],
        img(src="/images/sidebar/dyk-bottom.png")
    ]
except ImportError:
    class foo(object):
        __file__ = ''
    exponent = foo()
    footer = ''
    billboards = []
    anonymous_logo = "/images/divmod-org-logo.png"
    fileDir = None
    announce_sidebar = ''


class AnonymousBase(Base, RootCrudMixin, LoginMixin):
    ## This is terrible but easy way to order the tabs on the anon page
    billboards = billboards
    name = None
    def render_tabs(self, context, data):
        return ol(id="static-navigation")[
            a(href="/")[
                li[img(src=(self.name is None and sel or notsel) % 'Divmod')]
            ],
        [
            a(href=x)[
                li[img(src=(self.name == x and sel or notsel) % (x, ))]
            ]
            for x in self.billboards
        ]]

    def getDynamicChild(self, name, request):
        if name in self.billboards:
            return AnonymousBase(self.original).setBodyContent(name)
        elif fileDir and name in [x.split('.')[0] for x in os.listdir(fileDir)]:
            return AnonymousBase(self.original).setBodyContent(name)

    def setBodyContent(self, C):
        self.name = C
        self.body_content = xml(open(os.path.join(fileDir, "%s.html" % C)).read())
        return self

    def render_body(self, context, data):
        if hasattr(self.body_content, 'clone'):
            return self.body_content.clone()
        return self.body_content

    def render_sidebar(self, context, data):
        if self.sidebar_content:
            return self.sidebar_content.clone()

    def render_signup(self, context, data):
        from quotient.sysup import IControlPanel
        req = context.locate(inevow.IRequest)
        cp = IControlPanel(self.original.store)
        rv = []
        up = req.URLPath().click('/')
        httpsp = self.original.store.service.parent.httpsPort
        if httpsp == 443:
            httpss = ""
        else:
            httpss = ":" + str(httpsp)
        up.netloc = up.netloc.split(":")[0] + httpss
        up.scheme = 'https'
        https = str(up)
        signupurl = https+'signup'
        claimurl = https+'claim'
        if cp.actionAllowed(IControlPanel.USER_CREATION):
            rv.append(p[a(id='signup', href=signupurl)['Sign Up a New User']])
        if cp.actionAllowed(IControlPanel.TICKET_REDEMPTION):
            rv.append(p[a(id='claim', href=claimurl)['Claim a Ticket']])
        return rv

    def render_welcome(self, context, data):
        req = context.locate(inevow.IRequest)
        if req.args.has_key('login-failure'):
            return span(style="color: red")[ req.args['login-failure'] ]
        return h4['Log in.']

    def render_reset(self, context, data):
        return a(id='forgot', href="/reset/")[ "Forgot your password?" ]

    document = [
    xml('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n'),
    html(xmlns="http://www.w3.org/1999/xhtml", lang="en", **{'xml:lang': "en"})[
        head[
            title["Divmod"],
            link(rel="stylesheet", href="/css"),
        ],
        body[
            img(id="anonymous-logo", src=anonymous_logo),
            div(id="login-sidebar")[
                img(src="/images/sidebar/login-top.png"),
                div(id="login-content")[
                    div(render=directive('welcome')),
                    span(render=directive('loginAction')),
                ],
                img(src="/images/sidebar/login-bottom.png")
            ],
            div(id="sidebar")[
                img(src="/images/sidebar/dyk-top.png"),
                div(id="sidebar-content")[
                    div(render=directive('signup')),
                ],
                img(src="/images/sidebar/dyk-bottom.png")
            ],
            announce_sidebar,
            div(id="anonymous-main-body")[
                div(id="anonymous-main-body-content")[
                    render_body,
                ],
                div(id="anonymous-main-body-footer")[
                    footer
                ]
            ],
            div(id="anonymous-auxillary")[
                render_tabs,
##            ],
##            div(id="anonymous-help-login")[
##                a(href="Help")["Help"],
##                " | ",
##                a(href="/me/")["Login"]
            ]
        ]
    ]
]


#            div(render=directive('reset')),
#            div(render=directive('signup')),


class AnonymousRoot(AnonymousBase):
    body_content = invisible[
    invisible(render=directive('mainBillboard')),
]

    def __init__(self, *args, **kw):
        super(AnonymousRoot, self).__init__(*args, **kw)
        self.remember(self, inevow.ICanHandleException)
        self.remember(self, inevow.ICanHandleNotFound)

    def render_mainBillboard(self, context, data):
        if fileDir:
            flname = os.path.join(fileDir, "Divmod.html")
            if os.path.exists(flname):
                return xml(open(flname).read())
        return h1["Welcome to Quotient"], p["Quotient is your personal communications server for email, voip, and im."]

    def render_appname(self, context, data):
        ## don't do that dweeb
        return self.original.store.service.parent.parent.name

    def child_signup(self, request):
        return SignupPage(self.original)

    def child_claim(self, request):
        from quotient.web.powerup import DefaultPowerupDisplay
        from quotient.signup import IBoxOffice
        return ClaimPage(IBoxOffice(self.original.store))

    def child_users(self, request):
        from quotient.public import UsersPage
        return UsersPage(self.original)

    def child_exception(self, request):
        raise Exception("This is a test exception.")

    postLocation = ''

    storeID = ''

    def willHandle_notFound(self, request):
        return True

    def renderHTTP_notFound(self, request):
        """If a URL other than one we know how to deal with is requested
        by an anonymous user, we have no idea until they log in whether
        the URL should be 404 or not. So we just return the login page
        (for ANY url) and once the user logs in, we can then discover whether
        the URL is a 404 for that particular user or not.
        """
        return LoginOnly(self.original).renderHTTP(request)

    def willHandle_exception(self, request, failure):
        return True

    def renderHTTP_exception(self, request, reason):
        ## TODO: Write the exceptions to another, different log.
        log.err(reason)
        return SorryRenderer(self.original).renderHTTP(request)


def afterPostView(request, item):
    """For use in form-post methods.  After a post, view the given item.

    Useful for redirecting after a user logs in.
    """
    url = '/'+('/'.join(request.sitepath + ['me', str(item.storeID)]))
    request.setComponent(iformless.IRedirectAfterPost, url)

class SignupPage(Base):
    __implements__ = IUserSignup, Base.__implements__
    document = taglib.page("Divmod", anonymous=True)[
        webform.renderForms()
        ]

    def signup(self, request, username, password, realname):
        from quotient import tap2
        from quotient.prefup import IPrefs
        from quotient import sysup
        cp = sysup.IControlPanel(self.original.store)
        try:
            ava = sysup.makeUser(self.original.store, username, password, cp.POWERUPS, True)
        except sysup.SignupDisallowed, e:
            raise formless.InputError(str(e))
        else:
            cc = ICredentialsChecker(self.original.store)
            topDomain = cc.getDomain().name
            p = IPrefs(ava)
            p.setPref('realname', realname)
            p.setPref('replyto', '%s@%s' % (username, topDomain))
            s = request.getSession()
            res = inevow.IResource(ava)
            s.setResourceForPortal(res, s.guard.resource.portal, res.logout)
            afterPostView(request, p)
            request.setComponent(iformless.IRedirectAfterPost, "/me/%s" % p.storeID)

    postLocation = 'signup'

    storeID = ''

class ClaimPage(Base):
    document = taglib.page(title="Claim a Ticket", anonymous=True)[
        webform.renderForms()
        ]

class UserFrontPage(AnonymousRoot):
    def render_welcome(self, ctx, data):
        """If the user visits / while logged in, a bunch of links to things only available
        on the anonymous avatar IResource are going to be broken.
        Let us instead redirect them to /me/ automatically.
        """
        request = ctx.locate(inevow.IRequest)
        request.redirect("/me/")
        return ''

    def render_signup(self, context, data):
        return ''

    def render_reset(self, context, data):
        return ''


class NotRenderable(Base):
    document = html[head[title["Too Bad"]], body["I don't know how to render ", str, "."]]


class Me(Base):
    def __init__(self, *args, **kw):
        super(Me, self).__init__(*args, **kw)
        self.remember(self, inevow.ICanHandleException)
        self.remember(self, inevow.ICanHandleNotFound)

    adaptsAvatar = True

    def render_visible(self, context, data):
        from quotient import prefup
        from quotient.web import pools
        pl = prefup.IPrefs(self.original).getPrimaryNavPool()
        return invisible(data=pl)[pools.PoolBrowser(pl, suggestedView="Gallery")]

    def render_redirect(self, ctx, avatar):
        """Redirect the user from the /me/ page to the mail inbox.

        Remove this when we have some useful summary page on the /me/ url.
        """
        from quotient import application
        divmod = application.IDivmodApp(avatar)
        appnav = divmod.getApplicationNav()
        redirectTo = list(appnav)[0]
        request = ctx.locate(inevow.IRequest)
        redirectURL = str(url.URL.fromRequest(request).sibling(redirectTo.storeID))
        print "REDIRECT URL", redirectURL
        request.redirect(redirectURL)
        return ''

    document = taglib.page(title="Quotient")[
        render_redirect,
        "This is the page where we give you an overview of all the features including instructions"
        " on how to get started."
    ]

    def logout(self):
        print "LOGGED OUT"

    def getDynamicChild(self, name, request):
        if name == '':
            return self
        try:
            i = int(name)
            # get something from the db
            item = self.original.getItemByID(i)
            res = inevow.IResource(item,default=None)
            if res is None:
                return NotRenderable(item)
            else:
                return res

        except ValueError:
            pass

    def child_exception(self, request):
        raise Exception("This is a test exception.")

    def child_embeddedException(self, request):
        def render_exception(context, data):
            raise Exception("Here is an exception, horray")

        class ExcPage(Base):
            document = taglib.page(title="Woah")[
                div[
                render_exception
                ]
            ]
        return ExcPage(self.original)

    def willHandle_exception(self, request, failure):
        return True

    def renderHTTP_exception(self, request, reason):
        from quotient import prefup
        show = prefup.IPrefs(self.original).showInvisible
        if show:
            return FailureRenderer(reason).renderHTTP(request)
        log.err(reason)
        return SorryRenderer(self.original).renderHTTP(request)

    def willHandle_notFound(self, request):
        return True

    def renderHTTP_notFound(self, request):
        from quotient import prefup
        show = prefup.IPrefs(self.original).showInvisible
        return Quotient404(self.original, developer=show).renderHTTP(request)

    def renderHTTP(self, request):
        Base.renderHTTP(self, request)

class SorryRenderer(Base):
    document = taglib.page(title="Error", anonymous=True)[
        h1["Error"],
        "Sorry, an error occurred while trying to access the url ",
        render_url,
        ". The error has been logged. Try going:",
        ul[
            li[a(href="javascript:history.back(1)")["Back"]],
            li[a(href="/")["Home"]]
        ]
    ]


class FailureRenderer(rend.Page):
    def render_failure(self, context, data):
        from nevow import failure
        return failure.formatFailure(data)

    # don't try to render nav; probability of exception is high
    docFactory = loaders.stan(
    taglib.page(title="Error", failure=True)[
        render_failure
    ])


counter = itertools.count().next


def column(method, default=False, width=200, name=None):
    method.position = counter()
    method.default = default
    method.width = width
    indname = ' '.join(map(str.capitalize, method.__name__.split('_')[1:]))
    if name is None:
        name = indname
    method.indexname = indname
    method.colname = name


class TabularBase(components.Adapter, rend.RenderFactory):
    __implements__ = ITabularRenderer,

    invisible = property(lambda self: getattr(self.original, 'invisible', False))

    def render_received(self, context, data):
        from quotient import prefup
        dr = data.dateCreated
        # convert to datetime object
        local = tpython.mkgmtime(dr)
        dtobj = datetime.datetime.fromtimestamp(local, timezones.Local)
        # and here's where we have to respect the user's account setting, hoorj
        try:
            res = context.locate(inevow.IResource)
            astz = dtobj.astimezone(prefup.IPrefs(res.avatar).timezone)
        except components.CannotAdapt:
            astz = dtobj
        return astz.strftime('%Y/%m/%d %H:%M:%S')
    column(render_received, False, 100, 'Date')

    def render_name(self, context, data):
        if data.name is None:
            name = "%s (Unnamed)" % data.__class__.__name__
        else:
            name = taglib.headerToUTF8(data.name)
            # XXX: context.key.split('.')[0] is a total hack
        return taglib.href(data)(id='%s.%s' % (context.key.split('.')[0], name))[name]
    column(render_name, True, 300)

noThing = TabularBase(None)


def action_delete(self, pool):
    item = self.original
    ava = item.store
    item.touch()
    item.deletedFrom = []
    while item.pools:
        poolid = item.pools.keys()[0]
        item.deletedFrom.append(poolid)
        ava.getItemByID(poolid).removeItem(item)
    from quotient import gcup
    gcup.ICollector(ava).trash.addItem(item)


FourOhFour = rend.FourOhFour
