# default.py

from os.path import join, split, exists

from nevow import tags
from nevow import url
from formless import webform
from nevow import compy

from quotient import web
from quotient.application import IApplication
from quotient.copyright import version
from quotient.help import IHelpText

from twisted.python import log


def secondaryNav(context, data):
    from quotient.taglib import href
    appnav, selected = None, None
    if compy.implements(data, IApplication):
        appnav = data.getApplicationNav()
    elif hasattr(data, 'applicationRef'):
        appnav = data.applicationRef.getItem().getApplicationNav()
        selected = data
    else:
        app = compy.getAdapter(data, IApplication, default=None)
        if app != None:
            appnav = app.getApplicationNav()

##    log.msg("secnav type: %s nav: %s" % (type(data),str(appnav)))            

    if appnav is not None:
        return tags.ol(id="secondary-nav")[[
            href(x)(id=x is selected and "secondary-highlighted" or "SecNav.%s" % x.name)[ tags.li[x.name] ]
            for x in appnav
        ]]
    return tags.ol(id="secondary-nav")[
        tags.a[tags.li[
            tags.xml('&nbsp;'),
            tags.comment["Warning: We don't know how to render a secondary nav for object %r" % (data, )]
        ]]
    ]

helpfiles = join(split(web.__file__)[0], 'help')


def help(context, data):
    default = """
<p>Sorry, there is no help file available for this page. If you see this message, and you
required help you were unable to receive, please send an email to
<a href="mailto:support@divmod.org?subject=Help Requested for %s">support@divmod.org</a>.</p>
""" % data.__class__.__name__

    txt = compy.getAdapter(data, IHelpText, default=default)
##     if txt == default:
##         log.msg("need to write help for type: %s name: %s" % (type(data), getattr(data, 'name', 'No name attribute')))

    return [
        tags.div["Click anywhere to close help."],
        tags.xml(txt)]


from nevow import blocks


try:
    import exponent
    logged_in_logo = "/ximages/anonymous_logo.png"
    small_logo = '/ximages/small_logo.png'
except ImportError:
    logged_in_logo = "/images/divmod-org-logo.png"
    small_logo = "/images/divmod-org-logo-small.png"


def mainSkin(lib, title, body, anonymous, failure):
        if not anonymous and not failure:
## TODO Make these links aware of an application not installed on /
            nav = lib.navigation
        else:
            nav = ''

        if not failure:
            search = tags.form(style="display: inline;", render=lib.commandLineAction, key="form")[
                        tags.input(type="text", name="input", value="search", style="width: 100%"),
                        tags.input(type="hidden",name="_charset_"),
                    ]
        else:
            search = ''

        logo = tags.a(href=url.root)[tags.img(src=logged_in_logo, _class="logo", alt="Divmod Logo")]

        return [
            lib.startTimer,
            tags.xml('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n'),
            tags.html(xmlns="http://www.w3.org/1999/xhtml", lang="en", **{'xml:lang': "en"})[
        tags.head[
            tags.title[ title ],
            tags.link(rel="stylesheet", href="/freeform_css"),
            tags.link(rel="stylesheet", href="/css"),
            tags.link(rel="shortcut icon", href="/images/favicon.png"),
            tags.script(language="javascript", src="/scripts"),
            blocks.blocks_glue,
        ],
        tags.body[
            lib.pageclass,
            lib.errordebugging,
            tags.div(id="help-box", onclick="this.style.display='none';", style="display: none")[
                help
            ],
            tags.div(id="auxillary") [
                logo,
                #search,
                nav,
            ],
            tags.div(id="main-body")[
                tags.div(id="body-top")[
                    secondaryNav,
                    tags.div(id="recent-result")[
                        lib.recentResult
                    ],
                    tags.img(id="small-logo", src=small_logo)
                ],
                body,
                tags.div(style="text-align: right")[
                    tags.span(id='qvers')['Quotient v.', version],
                    lib.showTimer]
            ]
        ]
    ]
]


import os
from twisted.web import static
here = tuple(os.path.split(__file__)[:-1])
it = static.File(os.path.join(*here+ ('default.css',)))
css = lambda lib: it

imgPath = os.path.join(*here + ('defaultimages',))
imgs = static.File(imgPath)
images = lambda lib: imgs

missingPath = os.path.join(imgPath, 'bumnail.png')
M = static.File(missingPath)

missingImage = lambda lib: M

