#!/usr/bin/env python2.3
# -*- test-case-name: "quotient.test -xatop.test" -*-

# This module is part of the Quotient project and is Copyright 2003 Divmod:
# http://www.divmod.org/.  This is free software.  You can redistribute it
# and/or modify it under the terms of version 2.1 of the GNU Lesser General
# Public License as published by the Free Software Foundation.

import os, sys, glob
import os.path as osp
from os.path import join as opj

if not sys.version_info[:3] >= (2,3,0):
    print "Quotient requires Python 2.3.0 or greater."
    sys.exit(2)

from distutils.core import setup
from distutils.command.install_data import install_data
from distutils.command.sdist import sdist
from distutils.command.install import install

class install_data_twisted(install_data):
    def finalize_options (self):
        self.set_undefined_options('install',
            ('install_lib', 'install_dir')
        )
        install_data.finalize_options(self)

def callInDir(dir, f, *args, **kw):
    orig = os.getcwd()
    os.chdir(dir)
    try:
        f(*args, **kw)
    finally:
        os.chdir(orig)

class SdistExtra(sdist):
    user_options = sdist.user_options + [
        ('include-lupy=', None,
         "If specified, wget the version of Lupy specified as the "
         "argument and include it with this release."),
        ('include-pyndex=', None,
         "If specified, wget the version of Pyndex specified as the "
         "argument and include it with this release."),
        ('include-reverend=', None,
         "If specified, wget the version of Reverend specified as the "
         "argument and include it with this release."),
        ]

    reston_url = 'http://osdn.dl.sourceforge.net/sourceforge/'
    lupy_url = reston_url + 'lupy/Lupy-%s.tar.gz'
    reverend_url = reston_url + 'reverend/Reverend-%s.tar.gz'
    pyndex_url = reston_url + 'pyndex/Pyndex-%s.tar.gz'

    def initialize_options(self):
        sdist.initialize_options(self)
        self.include_lupy = self.include_pyndex = self.include_reverend = None

    def make_release_tree(self, base_dir, files):
        sdist.make_release_tree(self, base_dir, files)

        if not (self.include_lupy or self.include_pyndex or self.include_reverend):
            return

        def _(): #in directory base_dir
            os.mkdir('contrib')
            def _(): #in directory "contrib"

                # for any of pyndex, lupy, or reverend, download and
                # untar them if the user wants us to
                for p, url in ((self.include_pyndex, self.pyndex_url),
                               (self.include_lupy, self.lupy_url),
                               (self.include_reverend, self.reverend_url)):

                    # `p' will either be None or a version number, if the
                    # user passed --include-foo
                    if p:
                        url = url % p
                        print "Getting", url
                        fn = url.split('/')[-1]
                        #is the file there?
                        if not os.path.exists(fn):
                            os.system('wget ' + url)
                        #is the directory there?
                        if not os.path.exists(fn[:-len('.tar.gz')]):
                            os.system('tar xzf ' + fn)
                        os.system('rm ' + fn)
            callInDir('contrib', _)
        callInDir(base_dir, _)


class InstallExtra(install):
    def run(self):
        install.run(self)
        if not os.path.exists('contrib'):
            print "No contrib packages to install."
            return
        print "Regular install finished. Now installing contrib packages."

        # for any packages in our `contrib' directory, run their
        # setup.py with identical arguments to what were given to
        # _this_ setup.py
        for pdir in [x for x in os.listdir('contrib') if os.path.isdir('contrib/'+x)]:
            def _():
                ret = os.spawnvp(os.P_WAIT, sys.executable, [sys.executable, 'setup.py'] + sys.argv[1:])
            callInDir('contrib/' + pdir, _)

def filtrator(f):
    if not os.path.isfile(f):
        return 0
    if f[-3:].lower() == '.py':
        return 0
    if f[-4:].lower() in ('.pyc','.pyo'):
        return 0
    if f[0] == '.':
        return 0
    if '.svn' in f.split(os.sep):
        return 0
    #print 'okd',f,f.split(os.sep)
    return 1

templatePath = os.path.join('quotient','templates')
quotientPath = 'quotient'
quotientPlugins = [os.path.join('quotient', 'plugins.tml'),
                   os.path.join('quotient', 'components.wsv'),
                   os.path.join('quotient',  'webcomponents.wsv')]
skinPath = os.path.join('quotient', 'web', 'skins')
quotientSkin = [os.path.join(skinPath, 'default.css')]

datafiles = [
    (quotientPath, quotientPlugins),
    (skinPath, quotientSkin),
    ]

imageDataPath = os.path.join('quotient', 'web', 'skins', 'defaultimages')
webTemplatePath = os.path.join('quotient','web','templates')
billboardPath = os.path.join('quotient', 'web', 'billboards')
testDataPath = os.path.join('quotient', 'test', 'testdata')
asciiCorpusPath = os.path.join(testDataPath, 'ascii')
unicodeCorpusPath = os.path.join(testDataPath, 'unicode')

def pathAdder(arg, dirname, fnames):
    if dirname.lower() == '.svn':
        return
    toAdd = []
    for fname in fnames:
        fn = os.path.join(dirname, fname)
        if filtrator(fn):
            toAdd.append(fn)
    if toAdd:
        datafiles.append((dirname, toAdd))
os.path.walk(templatePath, pathAdder, 0)
os.path.walk(webTemplatePath, pathAdder, 0)
os.path.walk(testDataPath, pathAdder, 0)
os.path.walk(asciiCorpusPath, pathAdder, 0)
os.path.walk(unicodeCorpusPath, pathAdder, 0)
os.path.walk(imageDataPath, pathAdder, 0)
os.path.walk(billboardPath, pathAdder, 0)

setup(name='Quotient',
      version='0.9.1',
      description="Quotient - A conversation server",
      author="Amir Bakhtiar and the Divmod team",
      author_email="amir@divmod.org",
      url="http://www.divmod.org/",
      license='GNU LGPL',
      packages=['atop',
                'atop.test',
                'quotient',
                'quotient.functional',
                'quotient.proto',
                'quotient.smarts',
                'quotient.web',
                'quotient.web.skins',
                'quotient.scripts',
                'quotient.benchmarks',
                'quotient.test',
                'quotient.test.testdata',
                'quotient.test.testdata.ascii',
                'quotient.test.testdata.unicode'
                ],
      data_files=datafiles,
      cmdclass = {'install_data': install_data_twisted,
                  'sdist': SdistExtra,
                  'install': InstallExtra})
