
"""
Finds and prints out duplicate index keys.
"""

from atop import store

def findDuplicateSubStoreIndexKeys(cursor, _found=object()):
    currentIndex = None
    indexItems = {}
    for (key, value) in iter(cursor.next, None):
        parts = tuple(key.split(store.SEP))
        # Pool on the Store
        if len(parts) > 3 and parts[3] == 'index':
            if parts[:5] != currentIndex:
                # New Index
                currentIndex = parts[:5]
                indexItems = {}
            x = indexItems.get(value)
            if x is not None:
                if x is not _found:
                    yield x
                yield (parts, value)
                indexItems[value] = _found
            else:
                indexItems[value] = (parts, value)

def findDuplicateStoreIndexKeys(cursor, _found=object()):
    currentIndex = None
    indexItems = {}
    for (key, value) in iter(cursor.next, None):
        parts = tuple(key.split(store.SEP))
        # Pool on the Store
        if len(parts) > 2 and parts[2] == 'index':
            if parts[:4] != currentIndex:
                # New Index
                currentIndex = parts[:4]
                indexItems = {}
            x = indexItems.get(value)
            if x is not None:
                if x is not _found:
                    yield x
                yield (parts, value)
                indexItems[value] = _found
            else:
                indexItems[value] = (parts, value)

def main():
    env, db, lock = store.envdblock('../data/db')
    try:
        cursor = db.cursor()
        try:
            for (k, v) in findDuplicateStoreIndexKeys(cursor):
                print k, ',', int(v)
            for (k, v) in findDuplicateSubStoreIndexKeys(cursor):
                print k, ',', int(v)
        finally:
            cursor.close()
    finally:
        db.close()
        env.close()
        lock.unlock()

if __name__ == '__main__':
    main()
