#!/usr/bin/env python2.3

# system imports
import time, sys, os, signal


from twisted.protocols import irc
from twisted.internet import reactor, protocol, defer, task
from twisted.python import log, failure
from twisted.trial import util

defer.Deferred.debug = True

class PingProcessProtocol(protocol.ProcessProtocol):
    def __init__(self, d, numPings):
        """d is a deferred that will be called back when
        numPings requests are successfully received"""
        self.d = d
        self.numPings = numPings
        self.count = 0

    def outReceived(self, data):
##        log.msg('child process outRecieved: %s' % data)
        if "bytes from" in data and 'icmp_seq' in data:
            self.count += 1
##            log.msg("self.count: %s" % self.count)
            if self.count >= self.numPings:
                self.transport.loseConnection()
                self.killYourself()
##                log.msg('successful run')
                d, self.d = self.d, None
                d.callback('Hoorj!')

    def killYourself(self):
        pid = self.transport.pid
        if pid:
            os.kill(pid, signal.SIGKILL)
        
    def errReceived(self, data):
        # something using twisted.internet.app is screwing this up
        m = 'child process errRecieved: %s' % data
        log.msg(m)

##     def processEnded(self, status):
##         log.msg('childProcess processEnded with status: %s' % status)


class PingFailure(failure.DefaultException):
    pass

class Pinger(object):
    address = None
    name = None
    status = "OK/IDLE"
    deferred = None
    say = None

    def __init__(self, name, address, numPings=5, timeout=10, interval=180):
        (self.name, self.address, self.numPings,
         self.timeout, self.interval) = name, address, numPings, timeout, interval
        self.loop = task.LoopingCall(self.doPing)
        
    def startMonitoring(self):
##         msg = "starting periodic pinging of %s. will run every %s seconds" % (self.address,
##                                                                               self.interval)
##         self.say(msg)
        d = self.startLoop()
        self.status = "RUNNING"

        def weirdFailure(status):
            self.say(self.factory.channel,
                     'An unexpected situation has occured, please check my log file for details')
            self.status = "WTF?!"
            log.msg(status)

        d.addErrback(weirdFailure)

    def stopMonitoring(self):
        msg = "stopping periodic pinging of %s" % (self.address)
        log.msg('%s: %s' % (self.name, msg))
        self.stopLoop()

    def doPing(self):
        status = "RUNNING"
        d = defer.Deferred()
        ppp = PingProcessProtocol(d, self.numPings)
        reactor.spawnProcess(ppp, '/bin/ping', ['ping', self.address], usePTY=True)

        def onTimeout():
            ppp.killYourself()
            log.msg('failure in pinging %s' % self.address)
            self.status = "FAILED"
            self.say("ATTENTION! PING OF %s FAILED!!" % self.address)
            d.errback()

        self.delayed = reactor.callLater(self.timeout, onTimeout) 

        def success(ignore=None):
#            log.msg('success in pinging %s' % self.address)
            self.delayed.cancel()

            if self.status == "FAILED":
                log.msg('ping success after failure')
                self.say("WE'RE BACK! PING OF %s OK!" % self.address)

            self.status = "OK/IDLE"

        d.addCallback(success)

    def reportStatus(self):
        self.say("%s's monitor's status is %s" % (self.name, self.status))

    def startLoop(self):
        """do the ping every interval seconds"""
        d = self.deferred = self.loop.start(self.interval)
        return d
        
    def stopLoop(self):
        msg = 'stopping monitoring of %s, status: %s' % (self.address, self.status)
        log.msg(msg)
        
        self.say(msg)
        self.loop.stop()
    
    def nextRun(self):
        from twisted.python.runtime import seconds
        return self.loop.call.getTime() - seconds()
    


class PingBot(irc.IRCClient):
    """A nanny-bot"""
    servers = {'ohm': 'www.divmod.com', 'tesla': 'tesla.divmod.net',
               'erlang': 'erlang.divmod.net', 'faraday': 'faraday.divmod.net',
               'dummy': '192.168.2.2'}
    monitorOnStartup = ['ohm', 'tesla', 'erlang', 'faraday']
    numPings = 5
    timeout = 10
    interval = 180
    deferred = None
    
    def __init__(self):
        self.nickname = "pingbot"
        self.pingers = {}
        
    def connectionMade(self):
        irc.IRCClient.connectionMade(self)
        log.msg('connection made')

    def connectionLost(self, reason):
        irc.IRCClient.connectionLost(self, reason)
        log.msg('connection lost: %s', reason)

    # callbacks for events

    def signedOn(self):
        """Called when bot has succesfully signed on to server."""
        self.join(self.factory.channel)

    def joined(self, channel):
        """This will get called when the bot joins the channel."""
        log.msg("[I have joined %s]" % channel)
        for m in self.monitorOnStartup:
            self.cmd_startmonitoring(m, True)

    def privmsg(self, user, channel, msg):
        """This will get called when the bot receives a message."""
        user = user.split('!', 1)[0]
##        log.msg("<%s> %s" % (user, msg))
        if msg.startswith("%s:" % self.nickname):
            nick, cmd = msg.split(':')
            self.dispatch(cmd.strip())

    def dispatch(self, command):
        c = command.split()
        methName = "cmd_%s" % c[0].lower()

        args = []
        if len(c) > 1:
            args = c[1:]

 ##       log.msg('methName: %s\n args: %s' % (methName, args))
        try:
            meth = getattr(self, methName)#, self.cmd_usage)
            log.msg(meth)
            meth(*args)
        except AttributeError, e:
            log.msg(e)
            self.say("I didn't quite catch that")
            
    def say(self, msg):
        irc.IRCClient.say(self, self.factory.channel, msg)
            
    def cmd_usage(self, *ignored):
        reply = "hi! to see my commands type %s: commands" % self.nickname
        self.say(reply)

    def cmd_status(self, name=None):
        if name == 'all' or name == None:
            for p in self.pingers.itervalues():
                p.reportStatus()
        elif name in self.servers:
            if name in self.pingers:
                self.pingers[name].reportStatus()
            else:
                self.say('not monitoring %s' % name)
        else:
            self.say("i don't know anything about %s" % name)

    def cmd_startmonitoring(self, name, quiet=False):
##        log.msg('checking for server with nickname %s' % name)
        if name in self.servers and name not in self.pingers:
##            log.msg('found server %s with address %s' % (name, self.servers[name]))
            p = Pinger(name, self.servers[name], self.numPings, self.timeout, self.interval)
            p.say = self.say
            p.startMonitoring()
            self.pingers[name] = p
            if not quiet:
                self.say('okay, now monitoring %s' % name)
        elif name in self.pingers:
            self.say("already monitoring %s" % name)
        else:
            self.say("i don't know nothin' 'bout no %s" % name)

    def cmd_stopmonitoring(self, name):
        if name in self.servers:
            if name in self.pingers:
                self.pingers[name].stopMonitoring()
                del self.pingers[name]
            else:
                self.say("not monitoring %s" % name)
        else:
            self.say("don't know anything about %s" % name)

    def cmd_commands(self):
        msg = """startMonitoring: begin periodic pinging of %s
stopMonitoring: stop doing my thing
status: my current status, or status of the last run""" % self.address
        self.say(msg)
        


class PingBotFactory(protocol.ClientFactory):
    """A factory for PingBots.

    A new protocol instance will be created each time we connect to the server.
    """

    # the class of the protocol to build when new connection is made
    protocol = PingBot

    def __init__(self, channel):
        self.channel = channel

    def clientConnectionLost(self, connector, reason):
        """If we get disconnected, reconnect to server."""
        connector.connect()

    def clientConnectionFailed(self, connector, reason):
        print "connection failed:", reason
        reactor.stop()



from twisted.application import service, internet
from twisted.manhole import telnet


application = service.Application('PingNanny')
myService = service.IServiceCollection(application)
internet.TCPClient('irc.freenode.net', 6667, PingBotFactory('#divmod')).setServiceParent(application)

telnet.Shell.checkUserAndPass = lambda s, u, p: True    # yay! monkey-patch!
myFactory = telnet.ShellFactory()
myFactory.password = "guest"
myFactory.username = "guest"
internet.TCPServer(10001, myFactory).setServiceParent(application)

