#!/usr/bin/env python2.3
import sys

from twisted.web.domhelpers import findElementsWithAttribute as findewa
from twisted.web.domhelpers import findNodes
from twisted.web.domhelpers import findNodesNamed as findnn
from twisted.web.domhelpers import gatherTextNodes as gather
from twisted.web.microdom import parseXMLString

class Path(object):
    def __init__(self, action, path):
        self.action, self.path = action, path

    def __str__(self):
        return "%s\t%s" % (self.action, self.path)

    def __repr__(self):
        return "Path object at <%s>: %s" % (hex(id(self)), str(self))  

class LogEntry(object):
    _rev, _author, _date, _msg, _paths = None, None, None, None, None
    def __init__(self, domnode):
        self._node = domnode

    def __str__(self):
        s = ['-' * 79,
             'r%s | %s | %s ' % (self.revision, self.author, self.date),
             'Changed paths:',
             '\n'.join(['  ' + p.action + p.path for p in self.paths]),
             '\n',
             '%s\n' % self.msg]
        return '\n'.join(s)

    def _getRev(self):
        if not self._rev:
            self._rev = self._node.getAttribute('revision')
        return self._rev
    revision = property(_getRev)

    def _getAuthor(self):
        if not self._author:
            self._author = gather(findnn(self._node, 'author')[0])
            assert self._author
        return self._author
    author = property(_getAuthor)

    def _getDate(self):
        if not self._date:
            self._date = gather(findnn(self._node, 'date')[0])
        return self._date
    date = property(_getDate)
        
    def _getMsg(self):
        if not self._msg:
            self._msg = gather(findnn(self._node, 'msg')[0])
        return self._msg
    msg = property(_getMsg)

    def _getPaths(self):
        if not self._paths:
            paths = findnn(self._node, 'paths')[0]
            self._paths = [Path(n.getAttribute('action'), gather(n)) for n in findnn(paths, 'path')]
        return self._paths
    paths = property(_getPaths)

class Log(object):
    def __init__(self, xmlstring):
        self.xmlstring = xmlstring
        self.dom = dom = parseXMLString(xmlstring)
        self.log = log = findnn(dom, 'log')[0]   # should be only one
        self.entries = [LogEntry(le) for le in findnn(log, 'logentry')]

    def __getitem__(self, n):
        return self.entries[n]

    def __iter__(self):
        return iter(self.entries)

    def _revIter(self):
        def ri():
            for item in self.entries[::-1]:
                yield item
        return ri()
    revIter = property(_revIter)

if __name__  == "__main__":
    if len(sys.argv) < 2:
        stream = sys.stdin.read()
    else:
        stream = file(sys.argv[1], 'r').read()

    log = Log(stream)
    for e in log:
        print e 

